# =========================================================================== #
from typing import List

from pydantic import TypeAdapter

# --------------------------------------------------------------------------- #
from captura import fields
from captura.controllers.base import Data, ResolvedGrantDocument, ResolvedGrantUser
from captura.depends import DependsAccess, DependsCreate, DependsDelete, DependsUpdate
from captura.err import ErrAccessDocumentCannotRejectOwner, ErrAccessUser, ErrDetail
from captura.models import Level, LevelStr
from captura.schemas import (
    AsOutput,
    EventSchema,
    GrantCreateSchema,
    GrantSchema,
    OutputWithEvents,
    mwargs,
)
from captura.views import args
from captura.views.base import (
    BaseView,
    OpenApiResponseCommon,
    OpenApiResponseDocumentForbidden,
    OpenApiResponseUnauthorized,
    OpenApiTags,
)


class DocumentGrantView(BaseView):
    # NOTE: Updates should not be supported. It makes more sense to just delete
    #       the permissions and create new ones.
    view_routes = dict(
        delete_grants_document=dict(
            url="/{uuid_document}",
            name="Revoke Document Access for Users",
            # A true one off.
            responses={
                403: dict(
                    description="Cannot revoke grants of other owners.",
                    model=ErrDetail[ErrAccessDocumentCannotRejectOwner],
                )
            },
        ),
        post_grants_document=dict(
            url="/{uuid_document}",
            name="Invite Users to Access Document",
        ),
        get_grants_document=dict(
            url="/{uuid_document}",
            name="Read User Grants/Grant Requests for Document",
        ),
        patch_grants_document=dict(
            url="/{uuid_document}",
            name="Accept/Reject User Grant Requests for Document",
        ),
    )

    view_router_args = dict(
        tags=[OpenApiTags.grants],
        responses={
            **OpenApiResponseCommon,
            **OpenApiResponseUnauthorized,
            **OpenApiResponseDocumentForbidden,
        },
    )

    @classmethod
    def delete_grants_document(
        cls,
        delete: DependsDelete,
        uuid_document: args.PathUUIDDocument,
        uuid_user: args.QueryUUIDUser,
        *,
        pending: bool = False,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Revoke access to **document** for the **users** specified by
        `uuid_user`.
        """

        data: Data[ResolvedGrantDocument] = delete.access.d_grant_document(
            uuid_document,
            uuid_user,
            level=Level.view,
            exclude_deleted=not delete.force,
            pending=pending,
            exclude_pending=not pending,
        )

        events, grants = list(), list()
        if len(data.data.grants):
            delete.grant_document(data)
            grants = TypeAdapter(List[GrantSchema]).validate_python(
                data.data.grants.values()
            )
            data.commit(delete.session)
            events.append(EventSchema.model_validate(data.event))

        return mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=grants,
            events=events,
        )

    @classmethod
    def get_grants_document(
        cls,
        access: DependsAccess,
        uuid_document: args.PathUUIDDocument,
        uuid_user: args.QueryUUIDUserOptional = None,
        *,
        pending: bool = False,
        pending_from: args.QueryPendingFromOptional = None,
    ) -> AsOutput[List[GrantSchema]]:
        """List the active grants for a **document**. If `uuid_user` is
        specified, then filter results by these users; otherwise, return all
        of the grants for **document**.

        To view grant requests set `pending` to be `True`. It is important to
        note that access to the users is not checked, because, for instance,
        a private user might be assigned to the document.
        """

        data: Data[ResolvedGrantDocument] = access.d_grant_document(
            uuid_document,
            uuid_user,
            level=Level.view,
            pending=pending,
            exclude_pending=False,
        )
        grants = list(data.data.grants.values())
        if pending_from:
            pending_from_ = fields.PendingFrom[pending_from.name]
            grants = list(v for v in grants if v.pending_from == pending_from_)

        out = mwargs(
            AsOutput,
            data=TypeAdapter(List[GrantSchema]).validate_python(grants),
        )
        return out

    @classmethod
    def post_grants_document(
        cls,
        create: DependsCreate,
        uuid_document: args.PathUUIDDocument,
        uuid_user: args.QueryUUIDUser,
        *,
        level: args.QueryLevel = LevelStr.view,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Invite users to access a document by creating pending grants. Users
        will have to accept new grants.

        To request access to a document, use `POST /grants/users/<uuid>`.


        ### When `force=true`

        When grants exist, they are replaced and those invited must re-accept
        their grant to have any access.

        Eventually this will not be necessary. So for instance a user with
        `modify` permissions can be demoted to `view` permissions without
        having to accept another grant.


        ### When `force=false`

        When grants exist, they are not replaced or modified in any way and
        the non-existing grants are created. This will result in this endpoint
        being indempotent in this case and is why it is possibly to get single
        layered events out of this endpoint.
        """
        # NOTE: Level filters grants.
        fr_level = Level.resolve(level)
        data: Data[ResolvedGrantDocument] = create.access.d_grant_document(
            uuid_document,
            uuid_user,
            level=fr_level,
        )
        create.create_data = GrantCreateSchema(level=fr_level)
        data_final = create.grant_document(data)
        create.session.add_all(data_final.data.grants.values())
        data_final.commit(create.session)

        mm = mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=TypeAdapter(List[GrantSchema]).validate_python(
                data_final.data.grants.values()
            ),
            events=[EventSchema.model_validate(data_final.event)],
        )
        return mm

    @classmethod
    def patch_grants_document(
        cls,
        update: DependsUpdate,
        uuid_document: args.PathUUIDDocument,
        uuid_user: args.QueryUUIDUser,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Accept any grant requests *(likely made via
        `PATCH /grants/users/{uuid_user}`)*. To reject invitations use the
        corresponding ``Delete`` endpoint.

        Updating grants in place is not permitted. To replace a grant use
        `POST /grants/users/{uuid_user}` with `force=true`.
        """

        data: Data[ResolvedGrantDocument] = update.access.d_grant_document(
            uuid_document,
            uuid_user,
            level=Level.own,
        )
        update.grant_document(data)
        data.commit(update.session)
        assert data.event is not None

        out = mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=TypeAdapter(List[GrantSchema]).validate_python(
                data.data.grants.values()
            ),
            events=[EventSchema.model_validate(data.event)],
        )
        return out


class UserGrantView(BaseView):
    view_routes = dict(
        get_grants_user=dict(
            url="/{uuid_user}",
            name="Read User Access/Invitations to Documents",
        ),
        delete_grants_user=dict(
            url="/{uuid_user}",
            name="Revoke User Access to Documents",
        ),
        post_grants_user=dict(
            url="/{uuid_user}",
            name="Request User Access to Documents",
            response_description="New grants and their events.",
        ),
        patch_grants_user=dict(
            url="/{uuid_user}", name="Accept/Reject User Invitations to Documents"
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.grants],
        responses={
            **OpenApiResponseCommon,
            **OpenApiResponseUnauthorized,
            403: dict(
                model=ErrDetail[ErrAccessUser],
                description=(
                    "User is not the user specified by `uuid_user` or an " "admin."
                ),
            ),
        },
    )

    @classmethod
    def delete_grants_user(
        cls,
        delete: DependsDelete,
        uuid_user: args.PathUUIDUser,
        uuid_document: args.QueryUUIDDocument,
        *,
        pending: bool = False,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Revoke grants for a **user** for the provided **documents**. The
        intended use case is for users/admins to revoke their own grants.
        """

        data: Data[ResolvedGrantUser] = delete.access.d_grant_user(
            uuid_user,
            uuid_document,
            exclude_deleted=not delete.force,
            pending=pending,
            level=Level.view,
        )

        events, grants = list(), list()
        if len(data.data.documents):
            delete.grant_user(data)
            events.append(data.event)
            # NOTE: This must be done before the commit because the data
            #       becomes transient
            grants = TypeAdapter(List[GrantSchema]).validate_python(
                data.data.grants.values()
            )
            data.commit(delete.session)

        return mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=grants,
            events=[EventSchema.model_validate(ee) for ee in events],
        )

    @classmethod
    def get_grants_user(
        cls,
        access: DependsAccess,
        uuid_user: args.PathUUIDUser,
        uuid_document: args.QueryUUIDDocumentOptional = None,
        *,
        level: args.QueryLevel | None = None,
        pending: bool = False,
        pending_from: args.QueryPendingFromOptional = None,
    ) -> AsOutput[List[GrantSchema]]:
        """Get grants for a **user**. The indented use case is for users/admins
        to view the grants for any user. If **uuid_document** is not specified
        then all grants for the user are returned.

        To view grant requests set `pending` to be `True`.
        """

        data: Data[ResolvedGrantUser] = access.d_grant_user(
            uuid_user,
            uuid_document,
            pending=pending,
            level=level,
        )
        grants = data.data.grants
        if pending_from is not None:
            pending_from_ = fields.PendingFrom[pending_from.name]
            grants = {
                k: v for k, v in grants.items() if v.pending_from == pending_from_
            }

        return mwargs(
            AsOutput[List[GrantSchema]],
            data=TypeAdapter(List[GrantSchema]).validate_python(grants.values()),
        )

    @classmethod
    def patch_grants_user(
        cls,
        update: DependsUpdate,
        uuid_user: args.PathUUIDUser,
        uuid_document: args.QueryUUIDDocument,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Approve pending grants for user. The intended use case is for users
        to accept their own grants (this can also be done by admins).

        To invite users to a document you own, use
        `POST /grants/documents/{uuid}`.

        Note that grants cannot be updated in place. To replace/update a grant
        use `POST /grants/users/{uuid}` endpoint with `force=true`."""

        data: Data[ResolvedGrantUser] = update.access.d_grant_user(
            uuid_user,
            uuid_document,
            pending=True,
            level=Level.view,
        )

        update.grant_user(data)
        data.commit(update.session)
        assert data.event is not None

        out = mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=TypeAdapter(List[GrantSchema]).validate_python(
                data.data.grants.values()
            ),
            events=[EventSchema.model_validate(data.event)],
        )
        return out

    @classmethod
    def post_grants_user(
        cls,
        create: DependsCreate,
        uuid_user: args.PathUUIDUser,
        uuid_document: args.QueryUUIDDocument,
        *,
        level: args.QueryLevel,
    ) -> OutputWithEvents[List[GrantSchema]]:
        """Request **user** access to *public* **documents** specified by
        **uuid_document**. These grants will await approval from a document
        owner or admin view `PATCH /grants/documents/{uuid_document}`.

        This does not have `force` as users will likely never have a use case
        to drop their own **Level**.
        """

        # NOTE: Use does not have access to the requested documents typically.
        #       The only requirement to ask for access is to know the document
        #       uuid, which would be difficult to geuss in the case of.
        # NOTE: `token_user_grants` is usually the same as grants. But in this
        #       case they are not used. `grants` will be updated inside of
        #       `create.grant_user`.

        data: Data[ResolvedGrantUser] = create.access.d_grant_user(
            uuid_user,
            uuid_document,
            level=None,
        )
        create.create_data = GrantCreateSchema(level=Level[level.name])
        data_final = create.grant_user(data)
        create.session.add_all(data_final.data.grants.values())
        data_final.commit(create.session)

        grants = data_final.data.grants
        return mwargs(
            OutputWithEvents[List[GrantSchema]],
            data=TypeAdapter(List[GrantSchema]).validate_python(grants.values()),
            events=[EventSchema.model_validate(data_final.event)],
        )



# =========================================================================== #
from typing import Annotated, List

from fastapi import Body, Depends, HTTPException
from pydantic import TypeAdapter

# --------------------------------------------------------------------------- #
from captura import util
from captura.controllers.base import Data, ResolvedUser
from captura.depends import DependsAccess, DependsDelete, DependsRead, DependsUpdate
from captura.err import ErrAccessUser, ErrDetail
from captura.models import Collection, Document, User
from captura.schemas import (
    AsOutput,
    CollectionMetadataSchema,
    CollectionSearchSchema,
    DocumentMetadataSchema,
    DocumentSearchSchema,
    OutputWithEvents,
    UserExtraSchema,
    UserSchema,
    UserSearchSchema,
    UserUpdateSchema,
    mwargs,
)
from captura.views import args
from captura.views.base import (
    BaseView,
    OpenApiResponseCommon,
    OpenApiResponseUnauthorized,
    OpenApiTags,
)

logger = util.get_logger(__name__)

OpenApiResponseUser = {
    **OpenApiResponseCommon,
    403: {
        "model": ErrDetail[ErrAccessUser],
        "description": (
            "User must be logged in as the user specified in the url "
            "or an admin to not raise this status."
        ),
    },
}


def user(access: DependsAccess, uuid_user: args.PathUUIDUser) -> User:
    return access.user(uuid_user, exclude_public=True, return_data=False)


DependsUser = Annotated[User, Depends(user)]


class UserSearchView(BaseView):
    """Separate from :class:`UserView` so that it is clear that this is the
    sole batch of endpoints through which ``search`` style results are
    available. All other controllers should only support get by ``uuid``(s).
    """

    view_routes = dict(
        get_search_users=dict(
            name="Search Users",
            url="/{uuid_user}/users",
            #   ^^^^^^^^^^^^^^^^^^^^^  Search results ALWAYS scoped for users
            #   by uuid (since admins might need to simulate search results of
            #   other users).
        ),
        get_search_documents=dict(
            url="/{uuid_user}/documents",
            name="Search User Documents",
        ),
        get_search_collections=dict(
            url="/{uuid_user}/collections",
            name="Search User Collections",
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.users],
        responses=OpenApiResponseUser,
    )

    # ----------------------------------------------------------------------- #

    # NOTE: This includes the path parameter so that admins may simulate
    #       exactly the search results of another user.
    @classmethod
    def get_search_users(
        cls,
        user: DependsUser,
        read: DependsRead,
        param: Annotated[UserSearchSchema, Depends()],
    ) -> AsOutput[List[UserSchema]]:
        """Get user collaborators or just list some users."""

        res: List[User] = read.search_user(user, param)
        return mwargs(
            AsOutput[List[UserSchema]],
            data=TypeAdapter(List[UserSchema]).validate_python(res),
        )

    # TODO: Test that users cannot access private docs/colls of others here.
    @classmethod
    def get_search_documents(
        cls,
        user: DependsUser,
        read: DependsRead,
        param: DocumentSearchSchema = Depends(),
    ) -> AsOutput[List[DocumentMetadataSchema]]:
        res: List[Document] = read.search_user(user, param)
        return mwargs(
            AsOutput[List[DocumentMetadataSchema]],
            data=TypeAdapter(List[DocumentMetadataSchema]).validate_python(res),
        )

    @classmethod
    def get_search_collections(
        cls,
        uuid_user: args.PathUUIDUser,
        read: DependsRead,
        param: CollectionSearchSchema = Depends(),
    ) -> AsOutput[List[CollectionMetadataSchema]]:
        user: User = read.access.user(uuid_user, return_data=False)
        res: List[Collection] = read.search_user(user, param)
        return mwargs(
            AsOutput[List[CollectionMetadataSchema]],
            data=TypeAdapter(List[CollectionMetadataSchema]).validate_python(res),
        )


class UserView(BaseView):
    """Routes for user data and metadata.

    This will be put on ``/users``.
    """

    view_routes = dict(
        get_user=dict(
            url="/{uuid_user}",
            name="Read User",
        ),
        patch_user=dict(
            url="/{uuid_user}",
            name="Update User",
            responses=OpenApiResponseUnauthorized,
        ),
        delete_user=dict(
            url="/{uuid_user}",
            name="Delete User (and Associated Objects)",
            responses=OpenApiResponseUnauthorized,
        ),
    )
    view_router_args = dict(responses=OpenApiResponseUser, tags=[OpenApiTags.users])
    view_children = {"": UserSearchView}

    @classmethod
    def get_user(
        cls,
        uuid_user: args.PathUUIDUser,
        read: DependsRead,
    ) -> AsOutput[UserSchema]:
        """Get `user` metadata.

        For instance, this could be used to make a profile page.
        """

        user = UserExtraSchema.model_validate(read.access.user(uuid_user))
        return mwargs(AsOutput[UserSchema], data=user)

    @classmethod
    def patch_user(
        cls,
        uuid_user: args.PathUUIDUser,
        update: DependsUpdate,
        updates: Annotated[UserUpdateSchema, Body()],
    ) -> OutputWithEvents[UserSchema]:
        """Update the `user` specified by **uuid_user**."""

        update.update_data = updates
        data: Data[ResolvedUser] = update.a_user(
            uuid_user,
            resolve_user_token=update.token_user,
        )

        return mwargs(
            OutputWithEvents[UserSchema],
            data=UserSchema.model_validate(data.data.users[0]),
            events=[data.event],
        )

    @classmethod
    def delete_user(
        cls,
        uuid_user: args.PathUUIDUser,
        delete: DependsDelete,
        restore: bool = False,
    ) -> None:
        """Remove a user and their unshared documents and edits.

        Only the user themself or an admin should be able to call this
        endpoint.
        """

        data = delete.access.d_user(uuid_user)
        raise HTTPException(
            400,
            detail=(
                "User deactivation and deletion not yet supported."
                "(This is still a prototype)."
            ),
        )

        data = delete.a_user(uuid_user)
        data.commit(delete.session)
        # return mwargs(
        #     OutputWithEvents[UserSchema],
        #     events=[],
        #     data=UserSchema.model_validate(data.data.user[0]),
        # )



# =========================================================================== #
from datetime import datetime

import httpx
import typer

# --------------------------------------------------------------------------- #
from captura.fields import KindObject, Singular
from legere import flags
from legere.flags import Output
from legere.handlers import CONSOLE, ConsoleHandler
from legere.requests.base import BaseRequests, ContextData, params

__all__ = ("EventsRequests",)


class EventsRequests(BaseRequests):
    typer_check_verbage = False
    typer_commands = dict(
        prune="req_prune",
        read="req_read",
        search="req_search",
        restore="req_restore",
        objects="req_objects",
    )

    def req_callback(
        cls,
        output: flags.FlagOutput = Output.yaml,
        columns: flags.FlagColumns = list(),
    ):
        """Specify request output format."""
        cls.output = output
        if not columns:
            columns = ["api_origin", "timestamp", "uuid_root"]
            columns += ["uuid_parent", "uuid", "kind", "kind_obj"]
            columns += ["uuid_obj", "detail"]
            cls.columns = columns

        # TODO: Fix this. Looks like events cli is probably broken.
        cls.handler = ConsoleHandler(output=output, columns=columns, data=None)  # type: ignore

    @classmethod
    def req_prune(
        cls,
        _context: typer.Context,
        kind_obj: flags.ArgKindObject,
        uuid_obj: flags.ArgUUIDEvent,
        *,
        root: bool = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        p = params(root=root)
        if kind_obj == KindObject.event:
            return httpx.Request(
                "DELETE", f"/events/{uuid_obj}", headers=context.headers, params=p
            )
        return httpx.Request(
            "DELETE",
            context.url(f"/events/{Singular(kind_obj.name).name}/{uuid_obj}"),
            headers=context.headers,
            params=p,
        )

    @classmethod
    def req_read(
        cls,
        _context: typer.Context,
        kind_obj: flags.ArgKindObject,
        uuid_obj: flags.ArgUUIDEvent,
        *,
        root: bool = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        if kind_obj == KindObject.event:
            return httpx.Request(
                "GET",
                context.url(f"/events/{uuid_obj}"),
                headers=context.headers,
                params=dict(root=root),
            )

        # NOTE: Across many controllers.
        kind_object_plural = Singular(kind_obj.name).name
        return httpx.Request(
            "GET",
            context.url(f"/{kind_object_plural}/{uuid_obj}/events"),
            params=params(root=root),
            headers=context.headers,
        )

    @classmethod
    def req_search(
        cls,
        # recurse: flags.FlagKindRecurse = KindRecurse.depth_firsut,
        # flatten: flags.FlagFlatten = True,
        _context: typer.Context,
        *,
        kind: flags.FlagKind = None,
        kind_obj: flags.FlagKindObjectOptional = None,
        uuid_obj: flags.FlagUUIDEventObjectOptional = None,
        uuid_event: flags.FlagUUIDEventsOptional = None,
        before: flags.FlagBefore = None,
        after: flags.FlagAfter = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        res = httpx.Request(
            "GET",
            context.url("/events"),
            headers=context.headers,
            params=(
                params(
                    # flatten=flatten,
                    # recurse=recurse.value if recurse is not None else None,
                    before=datetime.timestamp(before) if before is not None else before,
                    after=datetime.timestamp(after) if after is not None else after,
                    uuid_event=uuid_event,
                    uuid_obj=uuid_obj,
                    kind=kind.value if kind is not None else None,
                    kind_obj=kind_obj.value if kind_obj is not None else None,
                )
            ),
        )
        return res

    @classmethod
    def req_restore(
        cls,
        _context: typer.Context,
        *,
        uuid_object: flags.FlagUUIDEventObjectOptional = None,
        uuid_event: flags.FlagUUIDEventOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        res = None
        match [uuid_object is None, uuid_event is None]:
            case [True, False]:
                res = httpx.Request(
                    "PATCH",
                    context.url(f"/events/{uuid_event}/objects"),
                    headers=context.headers,
                )
            case [False, True]:
                res = httpx.Request(
                    "PATCH",
                    context.url(f"/events/objects/{uuid_object}"),
                    headers=context.headers,
                )
            case [True, True] | [False | False] | _:
                CONSOLE.print(
                    "Exactly one of `--uuid-object` and `--uuid-event` must "
                    "be specified."
                )
                raise typer.Exit(1)

        return res

    @classmethod
    def req_objects(
        cls,
        _context: typer.Context,
        *,
        uuid_event: flags.ArgUUIDEvent,
        root: bool = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "GET",
            context.url(f"/events/{uuid_event}/objects"),
            params=params(root=root),
            headers=context.headers,
        )


if __name__ == "__main__":
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    events = typerize(EventsRequests)
    events()



# =========================================================================== #
import enum
import logging
import logging.config
import logging.handlers
import os
import secrets
from os import environ, path
from typing import Any, Type

import yaml
from rich.console import Console
from rich.syntax import Syntax
from sqlalchemy.orm import Session

LOG_LEVEL = logging.INFO

# =========================================================================== #
# PATH STUFF


class Path:
    @classmethod
    def ensure(cls, dirpath: str) -> None:
        if path.isfile(dirpath):
            raise ValueError(f"`{dirpath}` should not be a file.")

        if path.isdir(dirpath) or path.exists(dirpath):
            return

        os.mkdir(dirpath)

    @classmethod
    def base(cls, v: str) -> str:
        return path.join(PATH_BASE, v)

    @classmethod
    def app(cls, v: str) -> str:
        return path.join(PATH_APP, v)

    @classmethod
    def tests(cls, v: str) -> str:
        return path.join(PATH_TESTS, v)

    @classmethod
    def test_assets(cls, v: str) -> str:
        return path.join(PATH_TESTS_ASSETS, v)

    @classmethod
    def docker(cls, v: str) -> str:
        return path.join(PATH_DOCKER, v)

    @classmethod
    def config(cls, v: str) -> str:
        return path.join(PATH_CONFIG, v)

    @classmethod
    def plugins(cls, v: str) -> str:
        return path.join(PATH_PLUGINS, v)


PATH_BASE: str = path.realpath(path.join(path.dirname(__file__), "..", ".."))
PATH_APP: str = path.join(PATH_BASE, "src/app")
PATH_CLIENT: str = path.join(PATH_BASE, "src/client")
PATH_CONFIG: str = path.join(PATH_BASE, "configs")
PATH_DOCKER: str = path.join(PATH_BASE, "docker")
PATH_PLUGINS: str = path.join(PATH_BASE, "plugins")
PATH_TESTS: str = path.join(PATH_BASE, "tests")
PATH_TESTS_ASSETS: str = path.join(PATH_TESTS, "assets")


# =========================================================================== #
# Environment variables that have no configuration equivalent.


# ENV_DATA = []


def prefix_env(v: str) -> str:
    return f"{ENV_PREFIX}{v}"


def from_env(v: str, default: str | None = None, *, prefix: bool = True):
    envvar = prefix_env(v) if prefix else v
    w = environ.get(envvar, default)

    if w is None:
        msg = f"Could not determine value for environment variable `{w}`."
        raise ValueError(msg)
    return w


ENV_PREFIX = "CAPTURA_"
PATH_CONFIG_APP = from_env(
    "CONFIG_APP",
    Path.config("app.yaml"),
)
PATH_CONFIG_CLIENT = from_env(
    "CONFIG_CLIENT",
    Path.config("client.yaml"),
)
PATH_CONFIG_TEST_APP = from_env(
    "CONFIG_APP_TEST",
    Path.config("app.test.yaml"),
)
PATH_CONFIG_TEST_CLIENT = from_env(
    "CONFIG_CLIENT_TEST",
    Path.config("client.test.yaml"),
)
PATH_HOOKS: str = from_env(
    "HOOKS",
    path.join(PATH_BASE, "plugins/hooks.py"),
)
PATH_HOOKS_USE: bool = from_env("HOOKS_USE", "1") == "1"


PATH_STATIC = from_env("STATIC", Path.app("static"))
PATH_LOGS = from_env("LOGS", Path.base("logs"))
PATH_CONFIG_LOG = from_env("LOGS_CONFIG", Path.base("logging.yaml"))

PLUGINS_USE = from_env("PLUGINS_USE", "1") == "1"
VERBOSE = from_env("VERBOSE", "0") != "0"
VERBOSE_HTTPEXCEPTIONS = from_env("VERBOSE_HTTPEXCEPTIONS", "0") != "0"

# NOTE: Session secret ensures that middleware can reload the session properly.
SESSION_SECRET = from_env("SESSION_SECRET", secrets.token_urlsafe())


# =========================================================================== #
# Enum Stuff


def check_enum_opt_attr(
    cls: Type[Any],
    field: str,
    T_enum: Type[enum.Enum],
) -> None:
    if not hasattr(cls, field):
        msg = f"`{cls.__name__}` missing explicit `{field}`."
        raise AttributeError(msg)

    match getattr(cls, field, None):
        case T_enum() | None:  # type: ignore
            pass
        case bad:
            raise ValueError(
                f"`{cls.__name__}` has incorrect type for `{field}`."
                f"Expected `{T_enum}` or `None` (got `{bad}` of type "
                f"`{type(bad)}`)."
            )


# =========================================================================== sss
# LOGGING STUFF


# TODO: Add quehandler.
def setup_logging(config_path: str = PATH_CONFIG_LOG):
    with open(config_path, "r") as file:
        config = yaml.safe_load(file)

    logging.config.dictConfig(config)

    return config, logging.getLogger


DEFAULT_LOGGING_CONFIG, _get_logger = setup_logging()


def get_logger(name: str) -> logging.Logger:
    ll = _get_logger(name)
    return ll


CONSOLE_APP = Console()


def sql_render(session: Session, *qs) -> str:
    cmps = (
        f"{q.compile(session.bind, compile_kwargs=dict(literal_binds=True))};"
        for q in qs
    )
    cmp = "\n\n".join(cmps)
    return cmp


def sql(session: Session, *qs) -> None:
    cmp = sql_render(session, *qs)
    sql_syntax = Syntax(cmp, "mysql", theme="fruity", word_wrap=True)
    CONSOLE_APP.print(sql_syntax)



# =========================================================================== #

# --------------------------------------------------------------------------- #
from legere import ConfigCommands, DockerCommand
from legere.requests import Requests


def main():
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    client = typerize(Requests)
    client.add_typer(typerize(ConfigCommands, exclude_callback=True), name="config")
    client.add_typer(typerize(DockerCommand, exclude_callback=True), name="docker")
    client()


if __name__ == "__main__":
    main()



# =========================================================================== #
import hashlib
import secrets
import uuid
from datetime import datetime
from random import choice, randint
from typing import Any, Callable, Dict, Generic, List, Protocol, Set, Type, TypeVar

from faker import Faker
from faker.providers import internet
from sqlalchemy import Column, inspect

# --------------------------------------------------------------------------- #
from captura import __version__
from captura.fields import LENGTH_DESCRIPTION, LENGTH_NAME, LENGTH_URL
from captura.models import (
    Collection,
    Document,
    Event,
    Grant,
    KindEvent,
    KindObject,
    Level,
    PendingFrom,
    User,
)

fkit = Faker()
fkit.add_provider(internet)
_item_kind_event: List[KindEvent] = list(KindEvent)
_item_kind_obj: List[KindObject] = list(KindObject)
# _item_format: List[Format] = list(Format)
_item_pending_from: List[PendingFrom] = list(PendingFrom)
_item_level: List[Level] = list(Level)
_now = int(datetime.timestamp(datetime.now()))
_Mk: Dict[str, Callable[[], Any]] = dict(
    kind=lambda: choice(_item_kind_event),
    api_version=lambda: __version__,
    timestamp=lambda: randint(0, _now),
    uuid=lambda: secrets.token_urlsafe(8),
    name=lambda: fkit.text(LENGTH_NAME),
    description=lambda: fkit.text(LENGTH_DESCRIPTION),
    detail=lambda: fkit.text(LENGTH_DESCRIPTION),
    url=lambda: fkit.url()[:LENGTH_URL],
    url_image=lambda: fkit.url()[:LENGTH_URL],
    level=lambda: Level(randint(1, 3)),
    pending_from=lambda: choice(_item_pending_from),
    kind_obj=lambda: choice(_item_kind_obj),
    kind_event=lambda: choice(_item_kind_event),
    # format=lambda: choice(_item_format),
    # content=lambda: bytes(fkit.text(LENGTH_CONTENT), "utf-8"),
    # message=lambda: fkit.text(LENGTH_MESSAGE),
    admin=(mk_bool := lambda: bool(randint(0, 1))),
    deleted=mk_bool,
    public=mk_bool,
    pending=mk_bool,
    api_origin=lambda: "tests/dummy.py",
    content=lambda: dict(
        dummy=dict(
            utc_timestamp=datetime.utcnow().isoformat(),
            used_by=list(),
            tainted=False,
        ),
        tags=[],
    ),
    email=lambda: secrets.token_urlsafe(4) + "-" + fkit.email(),
)


def combos():
    while True:
        yield from (
            dict(
                level=level,
                pending_from=pending_from,
                deleted=deleted,
                pending=pending,
            )
            for level in (Level.view, Level.modify, Level.own)
            for pending_from in (PendingFrom.grantee, PendingFrom.granter)
            for deleted in (0, 1)
            for pending in (0, 1)
        )


def get_mk(column: Column):
    if column.name.startswith("uuid"):
        return  # _Mk["uuid"]
    elif column.name.startswith("id"):
        return
    elif column.name.startswith("_prototype"):
        return

    match column:
        case Column(primary_key=True):
            return
        case _:
            return _Mk[column.name]  # type: ignore


T_ResolvableContra = TypeVar(
    "T_ResolvableContra",
    Collection,
    User,
    Document,
    Grant,
    Event,
    covariant=True,
)


class MkDummyProvider(Protocol, Generic[T_ResolvableContra]):
    def __call__(self, **kwargs) -> T_ResolvableContra: ...


def create_mk_dummy(
    T_model: Type[T_ResolvableContra], exclude: Set[str] | None = None
) -> MkDummyProvider[T_ResolvableContra]:
    exclude = exclude or set()

    cols = {
        col.name: mk
        for col in inspect(T_model).columns
        if col.name not in exclude and (mk := get_mk(col)) is not None
    }

    def wrapper(**kwargs) -> T_ResolvableContra:
        base_kwargs = {key: mk() for key, mk in cols.items()}
        base_kwargs.update(kwargs)
        return T_model(**base_kwargs)

    return wrapper


class Mk:
    _user = staticmethod(create_mk_dummy(User, {"subject"}))

    @classmethod
    def user(cls):
        u = cls._user()
        u.uuid = str(uuid.uuid4())
        u.subject = hashlib.sha256(u.uuid.encode()).hexdigest()
        return u

    collection = staticmethod(create_mk_dummy(Collection))
    document = staticmethod(create_mk_dummy(Document))
    grant = staticmethod(create_mk_dummy(Grant))
    event = staticmethod(create_mk_dummy(Event))



# import asyncio
# import enum
# import json
# import secrets
# from typing import (Annotated, Any, AsyncGenerator, Dict, Generic, List, Self,
#                     Tuple, TypeVar)
#
# import httpx
# import typer
# import yaml
# from captura import __version__
# from captura.models import ChildrenAssignment, ChildrenGrant, Plural
# from captura.schemas import CollectionPostSchema, DocumentPostSchema, UserSchema
# from captura.views import AppView
# from pydantic import BaseModel, ConfigDict, Field, model_validator
#
# from legere import flags
# from legere.handlers import Handler
# from legere.requests import Requests
#
# from .requests import Requests
#
# # from pydantic.generics import GenericModel
#
# SpecResult = AsyncGenerator[httpx.Response | Tuple[httpx.Response], None]
#
#
# # NOTE: Distinguishes errors from `ApplyError` since cannot raise `typer.Exit`
# #       responsibly in many instances.
# class ApplyError(Exception):
#     @classmethod
#     def bad_mode(cls, bad: Any) -> Self:
#         msg = "expected one of `read`, `apply`, or `destroy`"
#         msg = f"Unknown mode {bad}, {msg}."
#         return cls(msg)
#
#     @classmethod
#     def cannot_match(cls, bad: Any) -> Self:
#         return cls(f"Cannot match `{bad}`.")
#
#
# class ApplyMode(enum.Enum):
#     read = "read"
#     apply = "apply"
#     destroy = "destroy"
#
#
# class ApplyState:
#     requests: Requests
#     mode: ApplyMode
#     handler: Handler
#
#     def __init__(self, requests: Requests, mode: ApplyMode, handler: Handler):
#         self.requests = requests
#         self.mode = mode
#         self.handler = handler
#
#
# # class ApplyStatus(BaseModel):
# #     uuid_status: Annotated[
# #         str,
# #         Field(default_factory=lambda: secrets.token_urlsafe(4)),
# #     ]
# #     uuid_status_previous: Annotated[str, Field()]
# #     uuid_obj: Annotated[str | None, Field(None)]
# #
# #     @classmethod
# #     def load(self, filepath: flags.ArgFilePath) -> str: ...
#
#
# class BaseAssoc(BaseModel):
#     # Collection/Document information.
#     kind_source: Annotated[ChildrenAssignment | None, Field(default=None)]
#     uuid_source: Annotated[str | None, Field(default=None)]
#
#     # Fill in one of these
#     name: Annotated[str | None, Field(default=None)]
#     uuid: Annotated[str | None, Field(default=None)]
#
#     @model_validator(mode="after")
#     def uuid_or_name(self) -> Self:
#         if (self.uuid is None) == (self.name is None):
#             raise ApplyError(
#                 "Specify exactly one of `uuid` and `name`. Got "
#                 f"`{self.uuid}` and `{self.name}`."
#             )
#
#         return self
#
#
# # class SpecGrant(BaseAssoc):
# #     async def resolve_uuid(self, state: ApplyState) -> str:
# #         if (name := self.name) is None:
# #             return self.uuid_source  # type: ignore
# #
# #         match self.kind_source:
# #             case ChildrenAssignment.documents:
# #                 res = await state.requests.documents.search(name_like=name)
# #             case ChildrenAssignment.collections:
# #                 res = await state.requests.collections.search(name_like=name)
# #             case _:
# #                 raise ApplyError("Unknown `kind_obj` `ChildrenAssignment`.")
# #
# #         if status := await state.handler.handle(res):
# #             raise typer.Exit(status)
# #         elif len(data := res.json()) != 1:
# #             raise ApplyError(f"Name `{name}` specifies many results.")
# #
# #         return data[0]["uuid"]
# #
# #     async def __call__(self, state: ApplyState) -> httpx.Response:
# #         client = state.requests.assignments
# #         documents, collections = client.documents, client.collections
# #         uuid_obj = [await self.resolve_uuid(state)]
# #
# #         if self.kind_source is None or self.uuid_source is None:
# #             raise ApplyError(
# #                 "`SpecAssignment` must define `kind_owner` and `uuid_owner` "
# #                 "to continue. These should either be specified directly or by"
# #                 "adding them from another called."
# #             )
# #
# #         match [state.mode, self.kind_source]:
# #             case ["apply", ChildrenAssignment.collections]:
# #                 res = await collections.create(self.uuid_source, uuid_obj)
# #             case ["apply", ChildrenAssignment.documents]:
# #                 res = await users.create(self.uuid_source, uuid_obj)
# #             case ["destroy", ChildrenAssignment.collections]:
# #                 res = await collections.delete(self.uuid_source, uuid_obj)
# #             case ["destroy", ChildrenAssignment.documents]:
# #                 res = await users.delete(self.uuid_source, uuid_obj)
# #             case [
# #                 _ as bad_mode,
# #                 ChildrenGrant.users | ChildrenGrant.documents,
# #             ]:
# #                 raise ApplyError.bad_mode(bad_mode)
# #             case ["apply" | "destroy", _ as bad]:
# #                 raise ApplyError(
# #                     f"Unknown source kind `{bad}` ` (expected one of "
# #                     "`documents` or `collection`)."
# #                 )
# #             case _ as bad:
# #                 raise ApplyError(f"Cannot match `{bad}`.")
# #
# #         if status := await state.handler.handle(res):
# #             raise typer.Exit(status)
# #         return res
#
#
# class SpecAssignment(BaseAssoc):
#     """For both :class:`CollectionApplySchema` and
#     :class:`DocumentApplySchema`.
#
#     :attr kind_source: The kind of object that the assignment is to be created
#         for. For instance, this could be `documents`.
#
#         This field should be set explicitly in YAML or by updating this object,
#         it is required for :meth:`__call__`.
#     :attr uuid_source: The uuid of the object that the assignment is to be
#         created for. This should be for an object of kind :attr:`kind_source`.
#
#         This field should be set explicitly in YAML or by updating this object,
#         it is required for :meth:`__call__`.
#     :attr name: Name of the target, e.g. if :attr:`kind_source` is
#         `collections` then this should be a document name.
#
#         Either this field or :attr:`uuid` is required.
#     :attr uuid: UUID of the target, e.g. if :attr:`kind_source` is
#         `collections` then this should be a document uuid.
#
#         Either this field or :attr:`uuid` is required.
#     """
#
#     async def resolve_uuid(
#         self, state: ApplyState
#     ) -> Tuple[httpx.Response | None, List[str]]:
#         if (name := self.name) is None:
#             return self.uuid_source  # type: ignore
#
#         match self.kind_source:
#             case ChildrenAssignment.documents:
#                 res = await state.requests.documents.search(name_like=name)
#             case ChildrenAssignment.collections:
#                 res = await state.requests.collections.search(name_like=name)
#             case _:
#                 raise ApplyError("Unknown `kind_obj` `ChildrenAssignment`.")
#
#         # if status := await state.handler.handle(res):
#         #     raise typer.Exit(status)
#         if len(data := res.json()) != 1:
#             raise ApplyError(f"Name `{name}` specifies many results.")
#
#         return res, [data[0]["uuid"]]
#
#     async def __call__(self, state: ApplyState) -> SpecResult:
#         client = state.requests.assignments
#         documents, collections = client.documents, client.collections
#
#         # NOTE: `res` will be returned to handler and checked there. Do not
#         #       check it there.
#         res, uuid_obj = await self.resolve_uuid(state)
#         if res is not None:
#             yield res
#
#         if self.kind_source is None or self.uuid_source is None:
#             raise ApplyError(
#                 "`SpecAssignment` must define `kind_owner` and `uuid_owner` "
#                 "to continue. These should either be specified directly or by"
#                 "adding them from another called."
#             )
#
#         match [state.mode, self.kind_source]:
#             case [ApplyMode.read, ChildrenAssignment.collections]:
#                 yield await collections.read(self.uuid_source, uuid_obj)
#             case [ApplyMode.read, ChildrenAssignment.documents]:
#                 yield await documents.read(self.uuid_source, uuid_obj)
#             case [ApplyMode.apply, ChildrenAssignment.collections]:
#                 yield await collections.create(self.uuid_source, uuid_obj)
#             case [ApplyMode.apply, ChildrenAssignment.documents]:
#                 yield await documents.create(self.uuid_source, uuid_obj)
#             case [ApplyMode.destroy, ChildrenAssignment.collections]:
#                 yield await collections.delete(self.uuid_source, uuid_obj)
#             case [ApplyMode.destroy, ChildrenAssignment.documents]:
#                 yield await documents.delete(self.uuid_source, uuid_obj)
#             case [
#                 _ as bad_mode,
#                 ChildrenAssignment.collections | ChildrenAssignment.documents,
#             ]:
#                 raise ApplyError.bad_mode(bad_mode)
#             case ["apply" | "destroy", _ as bad]:
#                 raise ApplyError(
#                     f"Unknown source kind `{bad}` ` (expected one of "
#                     "`documents` or `collection`)."
#                 )
#             case _ as bad:
#                 raise ApplyError.cannot_match(bad)
#
#         # await state.handler.handle(res)
#         # if status := await state.handler.handle(res):
#         #     raise typer.Exit(status)
#         return
#
#
# class SpecDocument(DocumentPostSchema):
#     collections: Annotated[
#         List[SpecAssignment] | None,
#         Field(default=None),
#     ]
#
#     # NOTE: Mode is considered by `SpecAssignment`.
#     async def handle_colllections(self, state: ApplyState) -> SpecResult:
#         if self.collections is None:
#             return
#
#         for collection in self.collections:
#             async for res in collection(state):
#                 yield res
#
#     async def __call__(self, state: ApplyState) -> SpecResult:
#         res, uuid_obj = await self.resolve_uuid(state)
#         if res is not None:
#             yield res
#
#         match state.mode:
#             case ApplyMode.apply:
#                 post = self.model_dump(exclude={"collections"})
#                 res = await state.requests.documents.create(post)
#                 yield res
#
#             case ApplyMode.destroy:
#                 # NOTE: Deletion should destroy assignments (to collections)
#                 uuid = await self.resolve_uuid(state)
#                 res = await state.requests.documents.delete(uuid)
#                 yield res
#
#             case _ as bad:
#                 raise ApplyError(f"Unknown mode `{bad}`.")
#
#         async for item in self.handle_colllections(state):
#             yield item
#
#     async def resolve_uuid(self, state: ApplyState) -> Tuple[httpx.Response, List[str]]:
#         # NOTE: Call only after `apply`.
#         res = await state.requests.documents.search(name_like=self.name)
#         if status := await state.handler.handle(res):
#             raise typer.Exit(status)
#
#         if len(data := res.json()) != 1:
#             raise ApplyError(f"Name `{self.name}` specifies many results.")
#
#         return res, [data[0]["uuid"]]
#
#
# class SpecCollection(CollectionPostSchema):
#     documents: Annotated[
#         List[SpecAssignment] | None,
#         Field(default=None),
#     ]
#
#     async def __call__(self, state: ApplyState) -> SpecResult:
#         match state.mode:
#             case ApplyMode.apply:
#                 post = self.model_dump(exclude={"collections"})
#                 res = await state.requests.collections.create(**post)
#                 yield res
#             case ApplyMode.destroy:
#                 # NOTE: Deletion should destroy assignments (to collections)
#                 uuid = await self.resolve_uuid(state)
#                 res = await state.requests.documents.delete(uuid)
#                 yield res
#             case _ as bad:
#                 raise ApplyError(f"Unknown mode `{bad}`.")
#
#     async def resolve_uuid(self, state: ApplyState) -> str:
#         # NOTE: Call only after `apply`.
#         res = await state.requests.collections.search(name_like=self.name)
#         if status := await state.handler.handle(res):
#             raise typer.Exit(status)
#
#         if len(data := res.json()) != 1:
#             raise ApplyError()
#
#         return data[0]["uuid"]
#
#
# class SpecUser(UserSchema):
#     documents: Annotated[List[SpecDocument] | None, Field(default=None)]
#     collections: Annotated[List[SpecCollection] | None, Field(default=None)]
#
#     async def __call__(self, state: ApplyState) -> SpecResult:
#         match state.mode:
#             case ApplyMode.apply:
#                 post = self.model_dump(exclude={"collections", "documents"})
#                 res = await state.requests.users.create(**post)
#                 yield res
#
#                 if status := await state.handler.handle(res):
#                     raise typer.Exit(status)
#
#                 if self.collections is not None:
#                     res = await asyncio.gather(
#                         collection(state) for collection in self.collections
#                     )
#                     yield res
#                     status = await state.handler.handle(res)
#                     if status:
#                         raise typer.Exit(status)
#                 if self.documents is not None:
#                     res = await asyncio.gather(
#                         document(state) for document in self.documents
#                     )
#                     yield res
#                     status = await state.handler.handle(res)
#                     if status:
#                         raise typer.Exit(status)
#             case ApplyMode.destroy:
#                 ...
#             case _:
#                 raise ApplyError()
#
#
# # NOTE: Singular naming has to be consistent with kind object.
# class Specs(enum.Enum):
#     documents = SpecDocument
#     collections = SpecCollection
#     users = SpecUser
#     assignments = SpecAssignment
#
#
# Spec = SpecUser | SpecCollection | SpecDocument | SpecAssignment
# T = TypeVar(
#     "T",
#     SpecUser,
#     SpecCollection,
#     SpecDocument,
#     SpecAssignment,
#     List[SpecUser],
#     List[SpecCollection],
#     List[SpecDocument],
#     List[SpecAssignment],
# )
#
#
# # NOTE: Holy fuck, this is awesome.
# class ObjectSchema(BaseModel, Generic[T]):
#     api_version: Annotated[str, Field(default=__version__)]
#     kind: Annotated[Plural, Field()]
#     spec: Annotated[T, Field()]
#
#     model_config = ConfigDict()
#
#     @model_validator(mode="before")
#     @classmethod
#     def validate_spec_type(cls, values: Any) -> Any:
#         value = values.get("spec")
#         if not isinstance(value, dict) and not isinstance(value, list):
#             return value
#
#         if (
#             (kind := values.get("kind")) is None
#             or isinstance(kind, Plural)
#             or kind == "list"
#         ):
#             return values
#         if kind in Plural.__members__:
#             kind = Plural.__members__[kind]
#         elif kind in Plural._value2member_map_:
#             kind = Plural._value2member_map_[kind]
#         else:
#             msg = str(tuple(Plural.__members__))
#             msg = f"Unknown object kind `{kind}`. Expected any of `{msg}`."
#             raise ValueError(msg)
#
#         TT = Specs[kind.value]
#         values.update(spec=TT.value.model_validate(value))
#         return values
#
#     # NOTE: Do not raise `typer.Exit` here since writing this such that data
#     #       classes do not contain typer. For the typerized function see
#     #       `ApplyMixins.apply`.
#     @classmethod
#     def load(cls, filepath: flags.ArgFilePath) -> Self:
#         with open(filepath, "r") as file:
#             data = yaml.safe_load(file)
#
#         if not isinstance(data, dict):
#             raise ApplyError(f"`{filepath}` must deserialize to a dictionary.")
#
#         return cls.model_validate(data)
#
#     async def __call__(self, state: ApplyState) -> AsyncGenerator[httpx.Response, None]:
#         match self.spec:
#             # case list() as items:
#             #     return await asyncio.gather(item(state) for item in items)
#             case (
#                 SpecUser()
#                 | SpecDocument()
#                 | SpecCollection()
#                 | SpecAssignment()
#                 # | SpecGrant()
#             ):
#                 return self.spec(state)
#             case _:
#                 raise ApplyError(f"No spec for `{type(self.spec)}`.")
#
#
# class ApplyMixins:
#     _state: ApplyState | None
#
#     @property
#     def state(self) -> ApplyState:
#         # SHOULD NOT EXIT GRACEFULLY. THIS HAPPENS WHEN BAD IMPL
#         if self._state is None:
#             raise ApplyError("State is missing.")
#         return self._state
#
#     # DECORATE THESE.
#     # async def read(
#     #     self,
#     #     filepath: flags.ArgFilePath,
#     # ) -> AsyncGenerator[httpx.Response, None]:
#     #     # self.state.mode = ApplyMode.read
#     #     # data = ObjectSchema.load(filepath)
#     #     # await data(self.state)
#     #
#     #     for item in "abcd1234":
#     #         await asyncio.sleep(1)
#     #         yield item
#     #
#     # async def apply(
#     #     self,
#     #     filepath: flags.ArgFilePath,
#     # ) -> AsyncGenerator[httpx.Response, None]:
#     #     data = self.load(filepath)
#     #     data(self.state)
#     #
#     # async def destroy(
#     #     self,
#     #     filepath: flags.ArgFilePath,
#     # ) -> AsyncGenerator[httpx.Response, None]:
#     #     data = self.load(filepath)
#     #     await data(self.state)
#
#
# async def apply(
#     bind,
#     state: ApplyState,
#     filepath: flags.ArgFilePath,
# ) -> AsyncGenerator[httpx.Response, None]:
#
#     app, config = None, state.requests.config
#     if not app:
#         app = AppView.view_router
#
#     async with httpx.AsyncClient(
#         base_url=config.host.remote,
#         app=app,
#     ) as client:
#         bind._client = client
#
#         # match self.state.mode:
#         #     case ApplyMode.read:
#         #         stream = self.read(filepath)
#         #     case ApplyMode.apply:
#         #         stream = self.apply(filepath)
#         #     case ApplyMode.destroy:
#         #         stream = self.destroy(filepath)
#         #
#         data = ObjectSchema.load(filepath)
#         stream = data(self.state)
#         async for thing in await stream:
#             yield thing



"""Api routers and functions. 

This includes a metaclass so that undecorated functions may be tested.

"""

# =========================================================================== #
import enum
import logging
from functools import cached_property
from http import HTTPMethod
from traceback import print_tb
from typing import (
    Annotated,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Literal,
    Self,
    Set,
    Tuple,
    Type,
    TypeAlias,
    TypeVar,
)

from fastapi import HTTPException
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    Tag,
    ValidationInfo,
    computed_field,
    field_validator,
)
from sqlalchemy.orm import Session

# --------------------------------------------------------------------------- #
from captura import util
from captura.auth import Token
from captura.fields import LevelHTTP, Singular
from captura.models import (
    AnyModel,
    Assignment,
    Base,
    Collection,
    Document,
    Event,
    Grant,
    KindObject,
    Level,
    ResolvableSingular,
    User,
    resolve_model,
    uuids,
)

logger = util.get_logger(__name__)
logger.level = logging.INFO


T_BaseController = TypeVar("T_BaseController", bound="BaseController")


class BaseController:
    """Base for :class:`Access`, :class:`Delete`, etc.

    :attr method: The ``http.HTTPMethod``. In some controllers this will be
        constant. In the access controller this will be used to check the
        levels of grants required to perform particular operations.
    :attr session: A ``Session``.
    :attr token: A ``Token``.
    """

    method: HTTPMethod
    session: Session
    _token: Token | None
    _token_user: User | None

    @property
    def level(self) -> Level:
        return LevelHTTP[self.method.name].value

    # TODO: Move  `token_user` and `user` to `Data`.
    @cached_property
    def token(self) -> Token:
        if self._token is None:
            raise HTTPException(401, detail="Token required.")
        return self._token

    @cached_property
    def token_user(self) -> User:
        if self._token_user is not None:
            return self._token_user
        token_user = self.token.validate_db(self.session)

        self._token_user = token_user
        return token_user

    # TODO: Rename this function.
    def then(self, type_: Type[T_BaseController], **kwargs) -> T_BaseController:
        "For chainable controllers."
        return type_(self.session, self._token, self.method, **kwargs)

    def token_user_or(
        self,
        resolve_user: ResolvableSingular[User] | None = None,
    ) -> User:
        return (
            User.resolve(self.session, resolve_user)
            if resolve_user is not None
            else self.token_user
        )

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
    ):
        self.session = session
        self._token_user = None

        match method:
            case str() if method in HTTPMethod._member_map_:
                self.method = HTTPMethod(method)
            case HTTPMethod():
                self.method = method
            case _:
                msg = f"Invalid input `{method}` for parameter `method`."
                raise ValueError(msg)

        match token:
            case dict() | None as raw:
                self._token = None if raw is None else Token.model_validate(raw)
            case Token() as token:
                self._token = token
            case bad:
                msg = f"Invalid input `{bad}` for parameter `token`."
                raise ValueError(msg)


def _uuid_set_from_model(
    data: Any,
    info: ValidationInfo,
    *,
    optional: bool = True,
) -> Set[str] | None:
    if data is not None:
        return data

    data_key = info.field_name
    if data_key is None:
        raise ValueError("WTF")

    data_key_source = data_key.replace("uuid_", "")
    if data_key_source not in info.data:
        raise ValueError(f"Invalid source `{data_key_source}` for uuids.")

    match info.data[data_key_source]:
        case None if optional:
            return None
        case tuple() as resolved:
            return uuids(resolved)  # type: ignore[type-var]
        case dict() as resovled:
            return uuids(tuple(resovled.values()))  # type: ignore[type-var]
        case Base() as other:
            return {other.uuid}
        case bad:
            return bad


def uuid_set_from_model(
    data: Any,
    info: ValidationInfo,
) -> Set[str]:
    res = _uuid_set_from_model(data, info)
    if res is None:
        raise ValueError()
    return res


def uuid_set_from_model_optional(
    data: Any,
    info: ValidationInfo,
) -> Set[str] | None:
    res = _uuid_set_from_model(data, info)
    return res


def uuid_from_model(data: Any, info: ValidationInfo) -> str:
    if data is not None:
        return data
    res = uuid_set_from_model(data, info)
    return res.pop()


UuidSetFromModelOptional = Annotated[
    Set[str] | None,
    Field(default=None, validate_default=True),
    BeforeValidator(uuid_set_from_model_optional),
]
UuidSetFromModel = Annotated[
    Set[str],
    Field(default=None, validate_default=True),
    BeforeValidator(uuid_set_from_model),
]
UuidFromModel = Annotated[
    str,
    Field(default=None, validate_default=True),
    BeforeValidator(uuid_from_model),
]


class KindData(str, enum.Enum):
    object_event = "object_event"
    event = "event"
    collection = "collection"
    document = "document"
    user = "user"
    edit = "edit"
    assignment_document = "assignment_document"
    assignment_collection = "assignment_collection"
    grant_document = "grant_document"
    grant_user = "grant_user"


ResolvableKindData: TypeAlias = KindData | KindObject | Tuple[KindObject, KindObject]


def resolve_kind_data(v: ResolvableKindData) -> KindData:
    match v:
        case KindData():
            return v
        case (
            KindObject.user
            | KindObject.collection
            | KindObject.edit
            | KindObject.document as kind
        ):
            return KindData(kind.name)
        case KindObject() as bad:
            msg = f"Cannot resolve scalar `KindObject` `{bad.name}`."
            raise ValueError(msg)
        case (KindObject.user, KindObject.document):
            return KindData.grant_user
        case (KindObject.document, KindObject.user):
            return KindData.grant_document
        case (KindObject.document, KindObject.collection):
            return KindData.assignment_document
        case (KindObject.collection, KindObject.document):
            return KindData.assignment_collection
        case (a, b):
            msg = f"Cannot resolve vector `KindObject` `({a.name}, {b.name})`."
            raise ValueError(msg)
        case _:
            raise ValueError(f"Cannot resolve `{v}`.")


def resolve_kind_assoc(v: KindData) -> Literal[KindObject.assignment, KindObject.grant]:
    match v:
        case KindData.grant_user | KindData.grant_document:
            return KindObject.grant
        case KindData.assignment_document | KindData.assignment_collection:
            return KindObject.assignment
        case bad:
            raise ValueError(f"Cannot resolve assoc for data kind `{bad}`.")


class BaseResolved(BaseModel):
    kind: ClassVar[KindData]
    registry: ClassVar[Dict[KindData, "Type[BaseResolved]"]] = dict()

    delete: bool = False
    _attr_name_targets: ClassVar[str]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if "Base" in cls.__name__:
            return

        util.check_enum_opt_attr(cls, "kind", KindData)
        if cls.kind in cls.registry:
            cls_existing = cls.registry[cls.kind].__name__
            raise ValueError(
                f"`registry` already has a resolved class `{cls_existing}` for"
                f"`{cls.kind}`. There should be exactly one class marked with "
                "any given kind."
            )
        cls.registry[cls.kind] = cls

    def register(self, session: Session) -> None: ...

    def refresh(self, session: Session) -> None: ...

    @classmethod
    def get(cls, resolvable_kind: ResolvableKindData) -> "Type[BaseResolved]":
        kind_data = resolve_kind_data(resolvable_kind)
        return BaseResolved.registry[kind_data]

    @classmethod
    def empty(cls, **kwargs_init) -> Self:
        "Yes, I know an ABC could be used here."
        raise ValueError("Not implemented!")


T_ResolvedPrimary = TypeVar("T_ResolvedPrimary", User, Collection, Document)


class BaseResolvedPrimary(BaseResolved):
    # ----------------------------------------------------------------------- #

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if "Base" in cls.__name__:
            return

        if cls.kind == KindData.object_event:
            cls._attr_name_targets = "events"
        else:
            cls._attr_name_targets = Singular(cls.kind.name).name

        # NOTE: How to check that a particular field is defined? `model_fields`
        #       is empty here but not when using the class otherwise. A
        #       solution without metaclasses is strongly preffered.
        #       This is nice to have because it verifies that
        #       ``_attr_name_targets`` is an actual field.

        # field_name = Singular(cls.kind.name).name
        # if field_name not in cls.model_fields:
        #     msg = f"`{cls.__name__}` should have field `{field_name}`. "
        #     msg += f"`{cls.model_fields}`."
        #     raise ValueError(msg)
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    @classmethod
    def empty(cls, **kwargs_init) -> Self:
        return cls(**{cls._attr_name_targets: tuple()}, **kwargs_init)

    # ----------------------------------------------------------------------- #

    def targets(self) -> Tuple[Any, ...]:
        return getattr(self, self._attr_name_targets)

    def err_nonempty(self) -> ValueError | None:
        # It is empty
        if not len(getattr(self, attr_name := self._attr_name_targets)):
            return None

        name = self.__class__.__name__
        msg = f"Primary data `{name}` should be empty (in field "
        msg += f"`{attr_name}`)."
        return ValueError(msg)

    def register(self, session: Session) -> None:
        """Because writing commit/refresh logic and wrapping to avoid breaking
        'all or none'.
        """

        if self.delete:
            return

        # NOTE: DO NOT COMMIT HERE! Commits occurs in `data.commit`
        targets = self.targets()
        session.add_all(targets)

    def refresh(self, session: Session) -> None:
        if self.delete:
            return

        for target in self.targets():
            session.refresh(target)


class BaseResolvedSecondary(BaseResolved):
    kind: ClassVar[KindData]
    kind_source: ClassVar[KindObject]
    kind_target: ClassVar[KindObject]
    kind_assoc: ClassVar[KindObject]
    _attr_name_source: ClassVar[str]
    _attr_name_assoc: ClassVar[str]
    _attr_name_target: ClassVar[str]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if "Base" in cls.__name__:
            return

        util.check_enum_opt_attr(cls, "kind_source", KindObject)
        util.check_enum_opt_attr(cls, "kind_target", KindObject)
        util.check_enum_opt_attr(cls, "kind_assoc", KindObject)
        cls._attr_name_source = cls.kind_source.name
        cls._attr_name_target = Singular(cls.kind_target.name).name
        cls._attr_name_assoc = Singular(cls.kind_assoc.name).name

    @classmethod
    def get_model(cls, for_: Literal["assoc", "target", "source"] = "assoc") -> Type:
        kind = getattr(cls, f"kind_{for_}")
        return resolve_model(kind)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def source(self) -> Document | User | Collection:
        return getattr(self, self._attr_name_source)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def target(self) -> Tuple[Collection, ...]:
        return getattr(self, self._attr_name_target)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def assoc(self) -> Dict[str, Any]:
        return getattr(self, self._attr_name_assoc)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def uuid_source(self) -> str:
        return getattr(self, "uuid_" + self._attr_name_source)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def uuid_target(self) -> Set[str]:
        return getattr(self, "uuid_" + self._attr_name_target)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def uuid_assoc(self) -> Set[str]:
        return getattr(self, "uuid_" + self._attr_name_assoc)

    def register(self, session: Session) -> None:
        """Because writing commit/refresh logic and wrapping to avoid breaking
        'all or none' of ACID.
        """
        # NOTE: Should not add targets or sources since they should not
        #       be modified in any of the functions that accept this `kind`
        #       of data. Instead, child data should be appended to data.

        if self.delete:
            return

        if self.assoc:
            session.add_all(self.assoc.values())

    def refresh(self, session: Session) -> None:
        if self.delete:
            return
        if not self.assoc:
            return

        for assoc in self.assoc.values():
            if assoc in session.deleted:
                continue
            session.refresh(assoc)


class ResolvedCollection(BaseResolvedPrimary):
    kind = KindData.collection

    collections: Tuple[Collection, ...]
    uuid_collections: UuidSetFromModel


class ResolvedDocument(BaseResolvedPrimary):
    kind = KindData.document

    documents: Tuple[Document, ...]
    uuid_documents: UuidSetFromModel

    token_user_grants: Dict[str, Grant]


class Resolved(BaseResolvedPrimary):
    kind = KindData.edit

    uuid_edits: UuidSetFromModel

    token_user_grants: Dict[str, Grant]


class ResolvedUser(BaseResolvedPrimary):
    kind = KindData.user

    users: Tuple[User, ...]
    uuid_users: UuidSetFromModel


class ResolvedEvent(BaseResolvedPrimary):
    kind = KindData.event

    events: Tuple[Event, ...]
    uuid_events: UuidSetFromModel


class ResolvedObjectEvents(ResolvedEvent):
    kind = KindData.object_event
    #
    # events: Tuple[Event, ...]
    # uuid_events: UuidSetFromModel

    obj: Annotated[AnyModel, Field()]
    uuid_obj: UuidFromModel
    kind_obj: Annotated[KindObject, Field()]


class ResolvedGrantUser(BaseResolvedSecondary):
    kind = KindData.grant_user
    kind_source = KindObject.user
    kind_target = KindObject.document
    kind_assoc = KindObject.grant

    user: User
    documents: Tuple[Document, ...]
    grants: Annotated[
        Dict[str, Grant],
        Field(
            description="A map from `document` uuids to their respective `grant`s for `user`."
        ),
    ]
    uuid_user: UuidFromModel
    uuid_documents: UuidSetFromModel
    uuid_grants: UuidSetFromModel

    # NOTE: See note inside of `Access.grant_user` about `token_user_grants`.
    token_user_grants: Dict[str, Grant]

    # TODO: Add TypedDict for these kwargs.
    # def grants(self, session: Session, **kwargs) -> Tuple[Grant, ...]:
    #     q = self.user.q_select_grants(self.uuid_documents, **kwargs)
    #     return tuple(session.execute(q).scalars())


class ResolvedGrantDocument(BaseResolvedSecondary):
    kind = KindData.grant_document
    kind_source = KindObject.document
    kind_target = KindObject.user
    kind_assoc = KindObject.grant

    document: Document
    users: Tuple[User, ...]
    grants: Annotated[
        Dict[str, Grant],
        Field(
            description="A map from `user` uuids to their respective `grant`s for `document`."
        ),
    ]
    uuid_document: UuidFromModel
    uuid_users: UuidSetFromModel
    uuid_grants: UuidSetFromModel

    # NOTE: See note inside of `Access.grant_document` about `token_user_grants`.
    token_user_grants: Dict[str, Grant]

    # TODO: Add TypedDict for these kwargs.
    # def grants(self, session: Session, **kwargs) -> Tuple[Grant, ...]:
    #     q = self.document.q_select_grants(self.uuid_users, **kwargs)
    #     return tuple(session.execute(q).scalars())


class ResolvedAssignmentCollection(BaseResolvedSecondary):
    kind = KindData.assignment_collection
    kind_source = KindObject.collection
    kind_target = KindObject.document
    kind_assoc = KindObject.assignment

    collection: Collection
    assignments: Dict[str, Assignment]
    documents: Tuple[Document, ...]
    uuid_collection: UuidFromModel
    uuid_assignments: UuidSetFromModelOptional
    uuid_documents: UuidSetFromModel

    # These are very helpful for handling cases in general


class ResolvedAssignmentDocument(BaseResolvedSecondary):
    kind = KindData.assignment_document
    kind_source = KindObject.document
    kind_target = KindObject.collection
    kind_assoc = KindObject.assignment

    document: Document
    assignments: Dict[str, Assignment]
    collections: Tuple[Collection, ...]
    uuid_document: UuidFromModel
    uuid_assignments: UuidSetFromModelOptional
    uuid_collections: UuidSetFromModel


T_Data = TypeVar(
    "T_Data",
    Annotated[ResolvedCollection, Tag("collection")],
    Annotated[ResolvedDocument, Tag("document")],
    Annotated[ResolvedUser, Tag("user")],
    Annotated[ResolvedAssignmentCollection, Tag("assignment_collection")],
    Annotated[ResolvedAssignmentDocument, Tag("assignment_document")],
    Annotated[ResolvedGrantUser, Tag("grant_user")],
    Annotated[ResolvedGrantDocument, Tag("grant_document")],
    Annotated[ResolvedEvent, Tag("event")],
    Annotated[ResolvedObjectEvents, Tag("object_events")],
)

kind_type_map = dict(
    user=User,
    collection=Collection,
    document=Document,
    grant_user=(User, Document),
    grant_document=(Document, User),
    assignment_collection=(Collection, Document),
    assignment_document=(Document, Collection),
)


class Data(BaseModel, Generic[T_Data]):
    """Because writing out complicated arguments is hard and composition will
    be simplified.

    :class:`Access` will resolve arguments which will be put into here. The
    other controller will then pass the data around so that signatures are
    not so nasty as `Access`, which will take in data, modify it appropriately
    (by performing CRUD and adding events to `Data` and return `Data`).

    This will allow any additional controllers to be sandwiched
    between with the sole requirement being that they match the following
    signature:

    .. code:: text

        (self, data: Data) -> Data

    For documentation of tagged unions, see
    https://docs.pydantic.dev/latest/concepts/unions/#discriminated-unions-with-callable-discriminator
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    data: T_Data
    event: Annotated[Event | None, Field(default=None)]
    token_user: Annotated[User | None, Field(default=None)]
    children: Annotated["List[Data]", Field(default_factory=list)]

    @field_validator("data", mode="before")
    def validate_raw(cls, v):
        return v

    @computed_field  # type: ignore[prop-decorator]
    @property
    def kind(self) -> KindData:
        return self.data.kind

    # @property
    # def resolved_cls(self) -> T_Data:
    #     return self.data.__class__

    # ----------------------------------------------------------------------- #

    def empty(self, kind_data: KindData | str, **kwargs_init) -> Self:
        "Make an empty version of self (same data, resolved is empty)."

        if isinstance(kind_data, str):
            kind_data = KindData[kind_data]

        T_Resolved = BaseResolved.registry[kind_data]
        mt = T_Resolved.empty(**kwargs_init)
        return self.__class__(
            data=mt,  # type: ignore
            event=self.event,
            token_user=self.token_user,
            children=list(),
        )

    def add(self, *items: "Data") -> None:
        for item in items:
            self.children.append(item)

    def register(self, session: Session) -> None:
        self.data.register(session)
        if self.event is not None:
            session.add(self.event)

        for child in self.children:
            child.register(session)

    def refresh(self, session: Session) -> None:
        # session.expire_all()
        self.data.refresh(session)
        if self.event is not None:
            session.refresh(self.event)
        for child in self.children:
            child.refresh(session)

    def commit(self, session: Session, commit: Any = None) -> None:
        if commit is not None:
            logger.warning("The `commit` argument is being removed.")
        self.register(session)
        try:
            session.commit()
        except Exception as err:
            print_tb(err.__traceback__)
            util.CONSOLE_APP.print("[red]" + str(err))
            session.rollback()
            raise HTTPException(500, "Database commit failure.")

        self.refresh(session)

    def types(self) -> Any:
        return kind_type_map[self.kind]  # type: ignore[return-type]


#
# T_OutputKind = TypeVar("T_OutputKind", AsOutput, OutputWithEvents)
# T_Output
#
# def primary_output(
#     data: Data[T_Data_Singular],
#     T_output_kind: T_OutputKind,
#     T_output: T_Output,
# ) -> Data[T_Data_


DataResolvedAssignment = (
    Data[ResolvedAssignmentCollection] | Data[ResolvedAssignmentDocument]
)
DataResolvedGrant = Data[ResolvedGrantUser] | Data[ResolvedGrantDocument]



# =========================================================================== #
from typing import Annotated

from fastapi import Body, Depends

# --------------------------------------------------------------------------- #
from captura.controllers.base import Data, ResolvedCollection
from captura.depends import DependsCreate, DependsDelete, DependsRead, DependsUpdate
from captura.err import ErrAccessCollection, ErrDetail
from captura.models import Collection
from captura.schemas import (
    AsOutput,
    CollectionCreateSchema,
    CollectionSchema,
    CollectionUpdateSchema,
    EventSchema,
    OutputWithEvents,
    mwargs,
)
from captura.views import args
from captura.views.base import (
    BaseView,
    OpenApiResponseCommon,
    OpenApiResponseUnauthorized,
    OpenApiTags,
)

# Depends


def collection(
    uuid_collection: args.PathUUIDCollection,
    read: DependsRead,
) -> Collection:
    collection: Collection = read.access.collection(uuid_collection, return_data=False)
    return collection


DependsCollection = Annotated[Collection, Depends(collection)]


# --------------------------------------------------------------------------- #

OpenApiResponseCollectionUnauthorized = {
    **OpenApiResponseUnauthorized,
    403: {
        "model": ErrDetail[ErrAccessCollection],
        "description": (
            "Raised when a token user does not own the specified collection "
            "(and in the case of `GET`, only when the collection is not "
            "public and the token user is not the collection owner)."
        ),
    },
}
OpenApiResponseCollection = {
    **OpenApiResponseCommon,
    **OpenApiResponseCollectionUnauthorized,
}


class CollectionSearchView(BaseView):
    view_routes = dict(
        # get_search_documents=dict(
        #     url="/{uuid_collection}/documents",
        #     # name="Search Collection Documents",
        # ),
        # get_search_collections="",
    )
    view_router_args = dict(
        tags=[OpenApiTags.collections],
        responses=OpenApiResponseCollection,
    )

    # @classmethod
    # def get_search_documents(
    #     cls,
    #     collection: DependsCollection,
    #     read: DependsRead,
    #     param: Annotated[DocumentSearchSchema, Depends()],
    # ) -> AsOutput[List[DocumentMetadataSchema]]:
    #     """Return metadata for `documents` (document `JSON` without `content`)
    #     in the `collection` specified by **uuid_collection** matching search
    #     params.
    #     """
    #
    #     res: Tuple[Collection, ...] = read.search_collection(collection, param)
    #     return mwargs(
    #         AsOutput[List[DocumentMetadataSchema]],
    #         data=TypeAdapter(List[DocumentMetadataSchema]).validate_python(res),
    #     )

    # @classmethod
    # def get_search_collections(
    #     cls,
    #     collection: DependsCollection,
    #     read: DependsRead,
    #     param: Annotated[CollectionSearchSchema, Depends()],
    # ) -> AsOutput[List[CollectionSchema]]:
    #     res: Tuple[Collection, ...] = read.search_user(collection, param)
    #     return mwargs(
    #         AsOutput[List[CollectionSchema]],
    #         data=TypeAdapter(List[CollectionSchema]).validate_python(
    #             res
    #         )
    #     )


class CollectionView(BaseView):
    """Routes for collection CRUD and metadata."""

    view_routes = dict(
        get_collection=dict(
            url="/{uuid_collection}",
            name="Read Collection",
            responses=OpenApiResponseCollectionUnauthorized,
        ),
        delete_collection=dict(
            url="/{uuid_collection}",
            name="Delete Collection",
            responses=OpenApiResponseCollectionUnauthorized,
        ),
        patch_collection=dict(
            url="/{uuid_collection}",
            name="Update Collection/Transfer Ownership",
            responses=OpenApiResponseCollectionUnauthorized,
        ),
        post_collection=dict(
            url="",
            name="Create a New Collection",
        ),
    )
    view_children = {"": CollectionSearchView}
    view_router_args = dict(
        tags=[OpenApiTags.collections],
        responses=OpenApiResponseCollection,
    )

    @classmethod
    def get_collection(
        cls, collection: DependsCollection
    ) -> AsOutput[CollectionSchema]:
        """Read a collection.

        To view documents in a collection use
        `GET /collections/{uuid_collection}/documents`.
        """
        return mwargs(
            AsOutput[CollectionSchema], data=CollectionSchema.model_validate(collection)
        )

    @classmethod
    def delete_collection(
        cls,
        delete: DependsDelete,
        uuid_collection: args.PathUUIDCollection,
    ) -> OutputWithEvents[CollectionSchema]:
        """Remove a `collection`.

        This will not remove the `documents` in a `collection` but will remove
        their respective `assignments`."""

        data: Data[ResolvedCollection] = delete.access.d_collection(
            uuid_collection, exclude_deleted=not delete.force
        )
        delete.collection(data)
        serial = CollectionSchema.model_validate(data.data.collections[0])

        data.commit(delete.session)
        return mwargs(
            OutputWithEvents[CollectionSchema],
            data=serial,
            events=[EventSchema.model_validate(data.event)],
        )

    @classmethod
    def patch_collection(
        cls,
        update: DependsUpdate,
        uuid_collection: args.PathUUIDCollection,
        updates: Annotated[CollectionUpdateSchema, Body()],
    ) -> OutputWithEvents[CollectionSchema]:
        """Update a collection/transfer collection ownership.

        When **uuid_user** is specified, the `collection` will be transfered to
        this user. Doing this will not change the `document` permissions such
        that the new owner of the `collection` will be able to access the
        `collection` and in fact the new owner will not see that such
        `documents` exist.
        """
        update.update_data = updates
        data: Data[ResolvedCollection] = update.a_collection(
            uuid_collection,
            resolve_user_token=update.token_user,
        )
        data.commit(update.session)

        return mwargs(
            OutputWithEvents[CollectionSchema],
            data=CollectionSchema.model_validate(*data.data.collections),
            events=[EventSchema.model_validate(data.event)],
        )

    @classmethod
    def post_collection(
        cls,
        create: DependsCreate,
        create_data: Annotated[CollectionCreateSchema, Body()],
    ) -> OutputWithEvents[CollectionSchema]:
        """Create a new collection.

        To add documents to this collection use
        `POST /assignments/collection/{uuid_collection}`.
        """

        # NOTE: User only needs a valid token to create a collection.
        create.create_data = create_data
        data = Data(
            token_user=create.token_user,
            event=None,
            data=ResolvedCollection.empty(),
            children=list(),
        )
        data_create = create.collection(data)
        data_create.commit(create.session)
        (collection,) = data_create.data.collections

        return mwargs(
            OutputWithEvents,
            events=[EventSchema.model_validate(data_create.event)],
            data=CollectionSchema.model_validate(collection),
        )



# =========================================================================== #
import importlib.util
import sys
from os import path
from types import ModuleType
from typing import Type

# --------------------------------------------------------------------------- #
from captura import util
from captura.views import AppView

logger = util.get_logger(__name__)


PATH_BASE: str = path.realpath(path.join(path.dirname(__file__), "..", ".."))


def get_hooks() -> None | ModuleType:
    if not util.PLUGINS_USE:
        return None

    if not path.isfile(util.PATH_HOOKS):
        logger.info(f"No hooks found at path `{util.PATH_HOOKS}`.")
        return None

    hooks_spec = importlib.util.spec_from_file_location("hooks", util.PATH_HOOKS)
    if hooks_spec is None:
        logger.info(f"No hooks found at path `{util.PATH_HOOKS}`.")
        return None

    logger.info("Found hooks.")
    hooks = importlib.util.module_from_spec(hooks_spec)
    sys.modules["hooks"] = hooks
    if hooks_spec.loader is None:
        raise ValueError(f"No loader for `{util.PATH_HOOKS}`.")

    hooks_spec.loader.exec_module(hooks)

    return hooks


def do_hooks(app_view: Type[AppView]):
    if not util.PLUGINS_USE:
        return

    if (module := get_hooks()) is None:
        return None

    if (captura_plugins_app := getattr(module, "captura_plugins_app")) is None:
        return None

    captura_plugins_app(app_view)



"""Database reports.

These were made because of dummies, but has uses in getting a 'big picture'
of the databases state. This will eventually be added as an extension to 
Captura.

Reports should be able to be saved, so this module defines 
:func:`create_tables` to create the optional reports tables.
"""

# =========================================================================== #
import enum
import itertools
import json
import secrets
from datetime import datetime
from typing import Annotated, Any, Dict, List, Tuple, TypeAlias

from fastapi import Depends, HTTPException
from pydantic import BaseModel, Field, TypeAdapter
from rich.align import Align
from rich.layout import Layout
from rich.panel import Panel
from rich.table import Table
from sqlalchemy import (
    JSON,
    Enum,
    ForeignKey,
    String,
    and_,
    func,
    literal_column,
    select,
    true,
    update,
)
from sqlalchemy.orm import Mapped, Session, mapped_column, relationship

# --------------------------------------------------------------------------- #
from captura import fields
from captura.depends import DependsSessionMaker
from captura.models import (
    Assignment,
    Base,
    Collection,
    Document,
    Event,
    Grant,
    MappedColumnUUID,
    User,
)
from captura.schemas import BaseSchema, TimespanLimitParams
from captura.views.base import BaseView


# NOTE: Schemas.
def row2dict(row):
    return {field: getattr(row, field) for field in row._fields}


class KindReport(str, enum.Enum):
    user = "user"
    aggregate = "aggregate"


ReportContentDataCountInt: TypeAlias = Annotated[int | None, Field(default=None)]


class ReportContentDataCount(BaseModel):
    users: ReportContentDataCountInt
    documents: ReportContentDataCountInt
    events: ReportContentDataCountInt
    collections: ReportContentDataCountInt
    grants: ReportContentDataCountInt
    assignments: ReportContentDataCountInt


class ReportContentData(BaseModel):
    count: ReportContentDataCount


class ReportContent(BaseModel):
    tags: Annotated[List[str], Field(default=list())]
    fn: Annotated[str | None, Field(default=None)]
    module: Annotated[str | None, Field(default=None)]
    data: ReportContentData


class ReportGrantSchema(BaseSchema):
    level: fields.FieldLevel
    pending_from: fields.FieldPendingFrom
    deleted: bool
    pending: bool
    count: int

    kind_mapped = None
    registry_exclude = True


class ReportGrantAggregateSchema(ReportGrantSchema):
    count_min: int | None
    count_max: int | None
    count_avg: float | None
    count_stddev: float | None


class BaseReportSchema(BaseSchema):
    note: str
    uuid_user: str | None
    uuid: str
    uuid_parent: str | None
    timestamp: datetime
    content: ReportContent

    kind_mapped = None
    registry_exclude = True

    def flatten(self) -> Dict[str, Any]:
        data = self.content.data
        count = data.count
        return dict(
            uuid_user=self.uuid_user,
            uuid=self.uuid,
            uuid_parent=self.uuid_parent,
            timestamp=self.timestamp,
            note=self.note,
            count_users=count.users,
            count_documents=count.documents,
            count_grants=count.grants,
            count_assignments=count.assignments,
            count_collections=count.collections,
            count_events=count.events,
        )

    def render_reports_grants(self) -> Panel:
        tt: Any
        if (reports_grants := getattr(self, "reports_grants", None)) is not None:
            tt = Table(title="Dummy Grants Report")
            cls = reports_grants[0].__class__

            for field in cls.model_fields:
                tt.add_column(field)

            for count, item in enumerate(reports_grants):
                style = "blue" if count % 2 else "cyan"
                if item.pending_from == fields.PendingFrom.created:
                    style = "on red"

                tt.add_row(
                    *(str(getattr(item, field)) for field in cls.model_fields),
                    style=style,
                )
        else:
            tt = "No grants to report."

        return Panel(
            Align.center(
                tt,
                vertical="middle",
            ),
        )

    def __rich__(self):
        if hasattr(self, "reports_grants"):
            panel_grants = self.render_reports_grants()
        else:
            panel_grants = None

        panel_report = self.render(align="left", exclude={"reports_grants", "user"})

        layout = Layout(visible=False)
        if panel_grants is not None:
            layout.split_row(
                Layout(panel_grants, name="left", visible=True),
                Layout(name="right"),
            )

        layout["right"].split_column(
            Layout(panel_report, name="top"),
            Layout(name="bottom", visible=False, ratio=3),
        )

        return layout


class ReportMinimalSchema(BaseReportSchema): ...


class ReportSchema(BaseReportSchema):
    reports_grants: List[ReportGrantAggregateSchema]


# --------------------------------------------------------------------------- #
# Models


class Report(Base):
    __tablename__ = "reports"

    uuid: Mapped[MappedColumnUUID] = mapped_column(primary_key=True)
    uuid_parent: Mapped[str] = mapped_column(
        ForeignKey(uuid, ondelete="CASCADE"),
        nullable=True,
    )

    # NOTE: For `snapshot all` because an aggregate report might have
    #       supporting user reports generated alongside it.
    children: Mapped[List["Report"]] = relationship(
        "Report",
        foreign_keys=uuid_parent,
    )

    timestamp: Mapped[int] = mapped_column(
        default=lambda: datetime.timestamp(datetime.now())
    )

    # NOTE: Is it worth caching? Does anybody care?
    # NOTE: See `content`.
    _content: Mapped[Dict[str, Any]] = mapped_column(
        JSON(),
        nullable=True,
        name="content",
        # default=functools.cache(lambda: ReportContent().model_dump()),
    )
    note: Mapped[str] = mapped_column(String(256))

    uuid_user: Mapped[str] = mapped_column(String(16), nullable=True)
    reports_grants = relationship(
        "ReportGrant", back_populates="report", passive_deletes=True
    )

    @property
    def content(self) -> ReportContent:
        return ReportContent.model_validate(
            self._content if self._content is not None else dict()
        )

    @content.setter
    def content(self, v: Any) -> None:
        data = ReportContent.model_validate(v)
        self._content = data.model_dump(mode="json")

    # NOTE: This is for some jupyter fun, so no schema provided since this
    #       will likely be throw into a dataframe.
    @classmethod
    def q_flat(cls, *additional_fields):
        return select(
            User.id.label("id_user"),
            cls.uuid.label("uuid"),
            cls.uuid_parent.label("uuid_parent"),
            cls.uuid_user.label("uuid_user"),
            cls.timestamp.label("timestamp"),
            cls.timestamp.label("note"),
            *(
                func.JSON_EXTRACT(cls.content, f"$.data.count.{item}").label(
                    f"count_{item}"
                )
                for item in fields.KindObject
            ),
            *additional_fields,
        ).join(User)

    # NOTE: Labeled like ReportContentDataCount
    @classmethod
    def q_content_data_count(cls, user: User | None = None):
        q_user = select(func.count(User.uuid))
        q_document = select(func.count(Document.uuid))
        q_collection = select(func.count(Collection.uuid))
        q_event = select(func.count(Event.uuid))
        q_grant = select(func.count(Grant.uuid))
        q_assignment = select(func.count(Assignment.uuid))

        if user is not None:
            q_document = select(Grant.id_document).where(
                Grant.pending_from == fields.PendingFrom.created,
                Grant.id_user == user.id,
            )
            q_document = select(func.count()).select_from(q_document.subquery())
            q_collection = q_collection.join(User).where(User.uuid == user.uuid)
            q_assignment = q_assignment.join(Collection).where(
                Collection.id_user == user.id
            )
            q_grant = q_grant.where(Grant.id_user == user.id)
            q_event = select(func.count(Event.uuid)).where(Event.uuid_user == user.uuid)

        return select(
            q_user.label("users"),
            q_document.label("documents"),
            q_collection.label("collections"),
            q_event.label("events"),
            q_grant.label("grants"),
            q_assignment.label("assignments"),
        )


class ReportGrant(Base):
    __tablename__ = "reports_grants"

    uuid: Mapped[MappedColumnUUID] = mapped_column(primary_key=True)
    uuid_report: Mapped[MappedColumnUUID] = mapped_column(
        ForeignKey("reports.uuid", ondelete="CASCADE"),
        nullable=False,
    )

    report = relationship("Report")

    level: Mapped[fields.Level] = mapped_column(Enum(fields.Level))
    pending: Mapped[bool] = mapped_column(default=True)
    pending_from: Mapped[fields.PendingFrom] = mapped_column(Enum(fields.PendingFrom))
    deleted: Mapped[bool]

    count: Mapped[int]
    count_avg: Mapped[float | None]
    count_stddev: Mapped[float | None]
    count_min: Mapped[int | None]
    count_max: Mapped[int | None]

    # NOTE: Should contain all of the columns specified in `ReportGrantSchema`.
    # NOTE: If mysl had `RETURNING` I'd use `INSERT INTO ... FROM SELECT ... RETURNING.
    @classmethod
    def q_select(cls, user: User | None = None):
        """Select the necessary data for a new grant."""

        _grant_agg = Grant.level, Grant.pending_from, Grant.deleted, Grant.pending
        q_reports_grants = (
            select(
                User.uuid.label("uuid_user"),
                Grant.level.label("level"),
                Grant.pending_from.label("pending_from"),
                Grant.pending.label("pending"),
                Grant.deleted.label("deleted"),
                func.count(Grant.uuid).label("count_per_user"),
            )
            .join(User)
            .group_by(User.uuid, *_grant_agg)
        )
        if user is not None:
            q_reports_grants = q_reports_grants.where(Grant.id_user == user.id)

        count_per_user: Any = literal_column("count_per_user")
        res_columns: Tuple[Any, ...] = (
            literal_column("level"),
            literal_column("pending_from"),
            literal_column("deleted"),
            literal_column("pending"),
        )

        q_grants = (
            select(
                *res_columns,
                func.min(count_per_user).label("count_min"),
                func.max(count_per_user).label("count_max"),
                func.avg(count_per_user).label("count_avg"),
                func.std(count_per_user).label("count_stddev"),
                func.sum(count_per_user).label("count"),
            )
            .select_from(q_reports_grants.subquery())
            .group_by(*res_columns)
            .order_by(*res_columns)
        )
        return q_grants


# =========================================================================== #
# Controllers


class ReportController:
    session: Session

    def __init__(self, session: Session):
        self.session = session

    def create(
        self,
        note: str,
        user: User | None = None,
        *,
        uuid_parent: str | None = None,
        fn: str | None = None,
        module: str | None = None,
        tags: List[str] | None = None,
    ) -> Report:

        session = self.session
        reports_grants = list(
            ReportGrant(**row2dict(row))
            for row in session.execute(ReportGrant.q_select(user=user)).all()
        )

        (content_data_count,) = (
            ReportContentDataCount(**row2dict(row))
            for row in session.execute(Report.q_content_data_count(user=user))
        )

        content = ReportContent(
            fn=fn if fn is not None else "ReportsContoller.create",
            module=module if module is not None else "dummy.reports",
            tags=tags if tags is not None else ["api", "aggregate"],
            data=ReportContentData(count=content_data_count),
        )
        report = Report(
            note=note,
            reports_grants=reports_grants,
            uuid_parent=uuid_parent,
        )
        report.content = content
        return report

    def create_aggregate(self, note: str, tags: List[str] | None = None) -> Report:
        return self.create(note, tags=tags)

    def create_user(
        self,
        note: str,
        user: User,
        tags: List[str] | None = None,
        *,
        uuid_parent: str | None = None,
    ) -> Report:
        return self.create(note, user, uuid_parent=uuid_parent, tags=tags)

    # NOTE: Trying to avoid using json items in reads.
    def read(
        self,
        params: TimespanLimitParams,
        uuid_user: str | None = None,
        exclude_children: bool = True,
        kind_count: fields.KindObject | None = None,
        kind_count_desc: bool = True,
        tags: List[str] | None = None,
    ) -> Tuple[Report, ...]:
        """Returns aggregate results be default. Otherwise returns user
        results when :param:`uuid_user` is not ``None``."""

        conds = []
        if params.after_timestamp is not None:
            conds.append(params.after_timestamp <= Report.timestamp)
        if params.before_timestamp is not None:
            conds.append(Report.timestamp <= params.before_timestamp)
        if uuid_user is not None:
            conds.append(Report.uuid_user == uuid_user)
        if exclude_children:
            conds.append(Report.uuid_parent.is_(None))
        if tags is not None:
            conds.append(
                func.JSON_OVERLAPS(
                    json.dumps(tags), func.JSON_VALUE(Report._content, "$.tags")
                )
            )

        q = select(Report).where(and_(*conds, true())).limit(params.limit)
        q = q.order_by(Report.timestamp.desc())
        if kind_count:
            _f = f"$.data.count.{kind_count.value}"
            f = func.JSON_EXTRACT(Report._content, _f)
            q = q.order_by(f.asc() if kind_count_desc else f.desc())

        # --------------------------------------------------------------------------- #
        from captura import util

        util.sql(self.session, q)

        return tuple(self.session.scalars(q))


# =========================================================================== #
# Views


def report_controller(sessionmaker: DependsSessionMaker):
    with sessionmaker() as session:
        yield ReportController(session)


DependsReportController = Annotated[
    ReportController,
    Depends(report_controller),
]


class ReportView(BaseView):
    # @classmethod
    # def serial(cls, user: User | None, report: Report):
    #     if user is not None:
    #         return ReportUserSchema.model_validate(report)
    #     else:
    #         return ReportAggregateSchema.model_validate(report)

    @classmethod
    def post_report_build(
        cls,
        report_controller: DependsReportController,
        note: str,
        uuid_user: str | None = None,
        return_report: bool = False,
        tags: List[str] | None = None,
    ) -> ReportSchema | None:
        session = report_controller.session
        user = None
        if uuid_user is not None:
            user = session.scalar(select(User).where(User.uuid == uuid_user))

        report = report_controller.create(note, user, tags=tags)
        session.add(report)
        session.add_all(report.reports_grants)
        session.commit()
        session.refresh(report)
        if not return_report:
            return None

        return ReportSchema.model_validate(report)

    @classmethod
    def post_report_build_all(
        cls,
        report_controller: DependsReportController,
        note: str,
        tags: List[str] | None = None,
    ) -> str:
        session = report_controller.session

        uuid_report = secrets.token_urlsafe(8)
        report = report_controller.create_aggregate(note, tags=tags)
        report.uuid = uuid_report
        report.children = list(
            report_controller.create_user(
                note, user, tags=tags, uuid_parent=uuid_report
            )
            for user in session.scalars(select(User))
        )
        session.add(report)
        session.commit()

        session.refresh(report)
        return report.uuid

    @classmethod
    def get_reports(
        cls,
        report_controller: DependsReportController,
        param: TimespanLimitParams,
        uuid_user: str | None = None,
        exclude_children: bool = True,
        kind_count: fields.KindObject | None = None,
        kind_count_desc: bool = True,
        tags: List[str] | None = None,
    ) -> List[ReportMinimalSchema]:
        reports = report_controller.read(
            param,
            uuid_user,
            exclude_children=exclude_children,
            kind_count=kind_count,
            kind_count_desc=kind_count_desc,
            tags=tags,
        )

        return TypeAdapter(List[ReportMinimalSchema]).validate_python(reports)

    @classmethod
    def delete_report(
        cls, report_controller: DependsReportController, uuid_report: str
    ):
        session = report_controller.session
        report = Report.if_exists(session, uuid_report)
        session.delete(report)
        session.commit()

        return None

    @classmethod
    def get_report(
        cls, report_controller: DependsReportController, uuid_report: str
    ) -> ReportSchema:
        session = report_controller.session
        report = Report.if_exists(session, uuid_report)
        return ReportSchema.model_validate(report)

    @classmethod
    def put_report(
        cls,
        report_controller: DependsReportController,
        uuid_report: str,
        note: str | None = None,
        tags: str | None = None,
    ) -> None:
        if note is None and tags is None:
            raise HTTPException(422, detail="One of `note` or `tags` required.")

        session = report_controller.session
        report = Report.if_exists(session, uuid_report)
        if note is not None:
            report.note = note
        if tags is not None:

            _tags_args = (("$.tags", tag) for tag in tags)
            tags_args = itertools.chain(*_tags_args)
            content = func.JSON_ARRAY_APPEND(Report._content, *tags_args)
            q = update(Report).values(_content=content)
            q = q.where(Report.uuid == uuid_report)
            session.execute(q)

        session.add(report)
        session.commit()



# =========================================================================== #
from typing import Type

# --------------------------------------------------------------------------- #
from captura import hooks, util
from legere.requests import Requests

logger = util.get_logger(__name__)


def do_hooks(client_command: Type[Requests]):
    if (module := hooks.get_hooks()) is None:
        return None

    if (captura_plugins_client := getattr(module, "captura_plugins_client")) is None:
        return None

    captura_plugins_client(client_command)



# =========================================================================== #
import base64
import enum
import json
import re
from hashlib import sha256
from os import path
from typing import Annotated, Any, Dict, Iterable, Self, Set, Tuple

import httpx
import jwt
from cryptography.hazmat.backends import default_backend

# from cryptography.hazmat.backends.openssl import _RSAPrivateKey, _RSAPublicKey
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from fastapi import HTTPException
from pydantic import BaseModel, Field, model_validator
from sqlalchemy import select
from sqlalchemy.orm import Session
from typing_extensions import Doc

# --------------------------------------------------------------------------- #
from captura import util
from captura.config import Config
from captura.models import User

# NOTE: Do not touch this! This pattern will have to run against many requests.
# TODO: Appearently fastapi has this pattern somewhere. Replace this with that.
logger = util.get_logger(__name__)
PATTERN_TOKEN: re.Pattern = re.compile(
    "^(?P<token>(?P<header>[\\w_-]+).(?P<payload>[\\w_-]+).(?P<signature>[\\w_-]+))$",
    flags=re.I,
)
PATTERN_BEARER: re.Pattern = re.compile(
    "^Bearer (?P<token>(?P<header>[\\w_-]+).(?P<payload>[\\w_-]+).(?P<signature>[\\w_-]+))$",
    flags=re.I,
)
PYTEST_KID = "000-000-000"
PATH_PYTEST_PUBLIC_KEY: str = util.Path.docker("pytest-public.pem")
PATH_PYTEST_PRIVATE_KEY: str = util.Path.docker("pytest-private.pem")

PrivateKey = Annotated[
    None | Any,
    Doc("RIP to type annotations here since the removal of `_RSAPublicKey`."),
]
PublicKey = Annotated[
    Dict[str, Any],
    Doc("RIP to type annotations here since the removal of `_RSAPublicKey`."),
]


# TODO: Add https://github.com/cak/secure. See https://github.com/auth0-developer-hub/api_fastapi_python_hello-world/tree/main for an example.
# https://github.com/auth0-developer-hub/api_fastapi_python_hello-world/blob/main/application/main.py
class Auth:
    """Authentication in general. Use :meth:`forAuth0` to create an instance
    that uses auth0 or use :meth:`forPyTest` to get a key pair for making
    tokens (which should only be necessary for integration tests).

    :attr private_key: The private_key, if it is provided.
    :attr public_keys: A set of public keys, probably from a JWKS.
    """

    config: Config
    issuer: str
    audience: str | Tuple[str, ...]
    private_key: PrivateKey
    public_keys: PublicKey

    def __init__(
        self,
        config: Config,
        public_keys: PublicKey,
        private_key: PrivateKey = None,
    ):
        # `issuer` and `audience` will not notice changes in config. Should not
        #  matter as config should be static.
        self.config = config
        self.issuer = self.config.auth0.issuer_url
        self.audience = self.config.auth0.api.audience
        self.private_key = private_key
        self.public_keys = public_keys

    @classmethod
    def forAuth0(cls, config):
        logger.debug("Getting Auth0 JWKS.")
        issuer = config.auth0.issuer
        result = httpx.get(f"https://{issuer}/.well-known/jwks.json")
        if result.status_code != 200:
            raise ValueError(f"Failed to fetch JWKS from issuer `{issuer}`.")

        jwks = result.json()

        # NOTE: I went through so much garbage documentation to find this in
        #       stack exchange post, lol.
        # https://stackoverflow.com/questions/68891213/how-to-decode-jwt-token-with-jwk-in-python
        public_keys = {}
        for jwk in jwks["keys"]:
            kid = jwk["kid"]
            _ = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
            public_keys[kid] = _

        return cls(config, public_keys)

    @classmethod
    def forPyTest(cls, config):
        # https://gist.github.com/gabrielfalcao/de82a468e62e73805c59af620904c124
        if not path.exists(PATH_PYTEST_PRIVATE_KEY) or not path.exists(
            PATH_PYTEST_PUBLIC_KEY
        ):
            logger.debug("Generating test authorization.")
            private_key = rsa.generate_private_key(
                public_exponent=65537, key_size=4096, backend=default_backend()
            )
            public_key = private_key.public_key()
            private_pem = private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
            public_pem = public_key.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            with (
                open(PATH_PYTEST_PUBLIC_KEY, "wb") as public_io,
                open(PATH_PYTEST_PRIVATE_KEY, "wb") as private_io,
            ):
                public_io.write(public_pem)
                private_io.write(private_pem)

        logger.debug("Loading pytest keypair.")
        with (
            open(PATH_PYTEST_PUBLIC_KEY, "rb") as public_io,
            open(PATH_PYTEST_PRIVATE_KEY, "rb") as private_io,
        ):
            public_key = serialization.load_pem_public_key(
                public_io.read(), backend=default_backend()
            )
            private_key = serialization.load_pem_private_key(
                private_io.read(), password=None, backend=default_backend()
            )

        logger.debug("Done loading!")
        return cls(config, {PYTEST_KID: public_key}, private_key)  # type: ignore

    def decode(self, raw: str, *, header: bool = True) -> Dict[str, Any]:
        """Get the key id from the JWT header and verify signature, audience,
        and issuer.

        :param raw: The raw ``authorization`` header content. This shold
            include ``Bearer``.
        :returns: The decoded JWT payload.
        """

        p = PATTERN_BEARER if header else PATTERN_TOKEN

        if (matched := p.match(raw)) is None:
            raise ValueError("Malformed JWT.")

        data_header_raw = matched.group("header") + "=="
        data_header_decoded = base64.b64decode(data_header_raw)
        data_header = json.loads(data_header_decoded)

        if (kid := data_header.get("kid")) is None:  # type: ignore
            raise ValueError("JWT header missing `kid`.")

        # NOTE: Hitting this line during tests with a remote host can be the
        #       result of the configuration of the remote host and not
        #       necessarily the tests.
        if (pk := self.public_keys.get(kid)) is None:
            detail = f"Malformed Token: JWT header contains unknown key id {kid}."
            raise HTTPException(401, detail=detail)

        return jwt.decode(
            matched.group("token"),
            pk,
            algorithms="RS256",  # type: ignore
            audience=self.audience,
            issuer=self.issuer,
        )

    def encode(self, payload) -> str:
        """When a private key is provided, use this to mint new tokens.

        This can be helpful in testing with pytest or messing around with cURL,

        :param payload: The JWT payload.
        """
        if self.private_key is None:
            raise ValueError("Cannot encode without a private key.")
        kid = list(self.public_keys.keys())[0]
        if "iss" not in payload:
            payload["iss"] = self.issuer
        if "aud" not in payload:
            payload["aud"] = self.audience
        return jwt.encode(
            payload, self.private_key, algorithm="RS256", headers={"kid": kid}
        )


# NOTE: Do not exit (via `HTTPException`) anywhere in this module besides in
#       in `try` prefixed methods.
def try_decode(
    auth: Auth,
    authorization: str,
    *,
    header: bool = True,
) -> Tuple[Dict[str, str] | None, HTTPException | None]:
    try:
        return auth.decode(authorization, header=header), None
    except jwt.DecodeError:
        _msg = "Failed to decode bearer token."
    except jwt.InvalidAudienceError:
        _msg = "Invalid bearer token audience."
    except jwt.InvalidIssuerError:
        _msg = "Invalid bearer token issuer."
    except jwt.InvalidTokenError as err:
        _msg = f"Invalid bearer token: {str(err)}"
    except ValueError as err:
        _msg = err.args[0]

    return None, HTTPException(401, detail="Invalid Token: " + _msg)


TokenPermissions = {
    "tier:free",
    "tier:paid",
    "tier:admin",
    "read:events",
    "read:reports",
}


class TokenPermissionTier(enum.Enum):
    """For now, only `admin` has an effect."""

    disabled = -10
    free = 0
    paid = 10
    admin = 50


class TokenPermissionRead(str, enum.Enum):
    """For reading special objects."""

    events = "events"
    reports = "reports"  # NOTE: Includes ``reports_grants``.


class TokenPermissionKind(enum.Enum):
    read = TokenPermissionRead
    tier = TokenPermissionTier


TokenPermissionPattern = re.compile("^(tier|read):([a-zA-Z]+)$")


def permission_parse(permissions: Iterable[str]) -> Dict[str, Any]:
    """Parse permissions array."""

    matched = {
        pp: mm
        for pp in permissions
        if (mm := TokenPermissionPattern.match(pp)) is not None
    }

    if len(bad := list(pp for pp in permissions if pp not in matched)):
        msg = "Raw JWT contains malformed permissions: `{}`."
        raise ValueError(msg.format(bad))

    parsed: Dict[str, Any] = dict()
    for mm in matched.values():
        key, value = mm.group(1), mm.group(2)
        parsed_value = parsed.get(key)
        match key:
            case "read":
                value_read = TokenPermissionRead(value)
                if parsed_value is None:
                    parsed["read"] = {value_read}
                else:
                    parsed["read"].add(value_read)
            case "tier":
                if parsed_value is not None:
                    msg = "`tier` can only be specified once."
                    raise ValueError(msg.format())
                parsed["tier"] = TokenPermissionTier[value]
            case _:
                raise ValueError()

    return parsed


class Token(BaseModel):
    # NOTE: uuid should not be in the token for the sake of maintainability.
    #       According to the standard [1], it should such that the ``sub``
    #       field of the jwt should be unique within the scope of the provider.
    #       The ``sub`` field is optional so it will necessary to first check
    #       that it exists.
    #
    # [1] https://www.rfc-editor.org/rfc/rfc7519#section-4.1

    subject: Annotated[str, Field(alias="sub")]
    tier: Annotated[
        TokenPermissionTier,
        Field(default=TokenPermissionTier.free),
    ]
    read: Annotated[
        Set[TokenPermissionRead],
        Field(default_factory=lambda: set()),
    ]

    @property
    def subject_256(self) -> str:
        return sha256(self.subject.encode()).hexdigest()

    @model_validator(mode="before")
    @classmethod
    def unfuck_permissions(cls, data: Any):
        if (pp := data.get("permissions")) is None:
            return data

        if "tier" in data or "read" in data:
            raise ValueError("`tier` and `read` should not be in raw data.")

        data.update(permission_parse(pp))
        return data

    def encode(self, auth: Auth) -> str:
        payload = self.model_dump(exclude={"subject", "tier", "read"})

        permissions = [f"read:{item.name}" for item in self.read]
        permissions.append(f"tier:{self.tier.name}")
        payload.update(permissions=permissions, sub=self.subject)

        return auth.encode(payload)

    # NOTE: This function is stupid. Why did I add it?
    def try_encode(self, auth: Auth) -> str:
        if auth.config.auth0.use:
            raise HTTPException(
                409,
                detail="Token minting is not available in auth0 mode.",
            )
        return self.encode(auth)

    @classmethod
    def decode(cls, auth: Auth, data: str, *, header: bool = True) -> Self:
        decoded = auth.decode(data, header=header)
        return cls.model_validate(decoded)

    # NOTE: This function is also stupid.
    @classmethod
    def try_decode(cls, auth: Auth, data: str, *, header: bool = True) -> Self:
        decoded, err = try_decode(auth, data, header=header)
        if err is not None:
            raise err

        return cls.model_validate(decoded)

    def validate_db(self, session: Session) -> User:
        q_user = select(User).where(User.subject == self.subject_256)
        user = session.execute(q_user).scalar()

        if user is None:
            raise HTTPException(
                401,
                detail=dict(
                    msg="User does not exist.",
                    subject=self.subject_256,
                ),
            )

        if user.admin and not self.tier == TokenPermissionTier.admin:
            raise HTTPException(
                401,
                detail=dict(
                    msg="Admin status inconsistent with database.",
                    subject=self.subject,
                    admin_user=user.admin,
                ),
            )

        return user



# =========================================================================== #
import enum
import json
import pathlib
from datetime import datetime
from typing import Annotated, List, Optional, TypeAlias

import typer
from typing_extensions import Doc

# --------------------------------------------------------------------------- #
from captura.fields import PendingFromStr
from captura.models import (
    ChildrenCollection,
    ChildrenDocument,
    ChildrenUser,
    KindEvent,
    KindObject,
    KindRecurse,
    LevelStr,
)


class Verbage(str, enum.Enum):
    read = "read"
    # TODO: Depricate the two bellow.
    read_document = "read_document"
    read_user = "read_user"

    search = "search"
    restore = "restore"
    update = "update"
    delete = "delete"
    create = "create"
    activate = "activate"

    # apply = "apply"
    # destroy = "destroy"


class Output(str, enum.Enum):
    raw = "raw"
    json = "json"
    yaml = "yaml"
    table = "table"


# --------------------------------------------------------------------------- #
# UUID Flags and Arguments
# NOTE: Annotations should eventually include help.

FlagOutput: TypeAlias = Annotated[Output, typer.Option("--output", "-o")]
FlagUUIDs: TypeAlias = Annotated[Optional[List[str]], typer.Option("--uuid")]


# User

ArgUUIDUser: TypeAlias = Annotated[str, typer.Argument(help="User uuid.")]
FlagUUIDUserOptional: TypeAlias = Annotated[
    Optional[str],
    typer.Option(
        "--uuid-user",
        help="Transfer ownship of a collection to this UUID.",
    ),
]
FlagUUIDUsers: TypeAlias = Annotated[
    List[str],
    typer.Option(
        "--uuid-user",
        help="A required list of user UUIDs.",
    ),
]
FlagUUIDUsersOptional: TypeAlias = Annotated[
    Optional[List[str]],
    typer.Option(
        "--uuid-user",
        help="An optional list of user UUIDs.",
    ),
]

FlagInvitationCode: TypeAlias = Annotated[
    str,
    typer.Option(
        "--invitation-code",
        help="The invitation code to approve/verify a demo user with.",
    ),
]
FlagInvitationEmail: TypeAlias = Annotated[
    str,
    typer.Option(
        "--invitation-email",
        help="The email corresponding the invitation code.",
    ),
]
FlagInvitationCodesOptional: TypeAlias = Annotated[
    Optional[List[str]],
    typer.Option(
        "--invitation-code",
        help="Invitation codes to search by.",
    ),
]
FlagInvitationEmailsOptional: TypeAlias = Annotated[
    Optional[List[str]],
    typer.Option(
        "--invitation-email",
        help="Invitation emails to search by.",
    ),
]


# --------------------------------------------------------------------------- #
# Documents

FlagUUIDDocuments: TypeAlias = Annotated[
    List[str],
    typer.Option("--uuid-document", help="Document uuids."),
]
FlagUUIDDocumentsOptional: TypeAlias = Annotated[
    Optional[List[str]],
    typer.Option(
        "--uuid-document",
        help="Optional document uuids.",
    ),
]
ArgUUIDDocument: TypeAlias = Annotated[str, typer.Argument()]


# --------------------------------------------------------------------------- #
# Collections

FlagUUIDCollections: TypeAlias = Annotated[List[str], typer.Option("--uuid-collection")]
FlagUUIDCollectionsOptional: TypeAlias = Annotated[
    List[str], typer.Option("--uuid-collection")
]
ArgUUIDCollection: TypeAlias = Annotated[str, typer.Argument()]


# --------------------------------------------------------------------------- #
# Events

FlagUUIDEvent = Annotated[str, typer.Option("--uuid-event")]
FlagUUIDEventOptional = Annotated[Optional[str], typer.Option("--uuid-event")]
FlagUUIDEvents = Annotated[List[str], typer.Option("--uuid-event")]
FlagUUIDEventsOptional = Annotated[Optional[List[str]], typer.Option("--uuid-event")]


# --------------------------------------------------------------------------- #
# Field flags

FlagColumns: TypeAlias = Annotated[List[str], typer.Option("--column")]
FlagLevel: TypeAlias = Annotated[LevelStr, typer.Option("--level")]

FlagName = Annotated[str, typer.Option("--name")]
FlagNameOptional = Annotated[Optional[str], typer.Option("--name")]

FlagDescription = Annotated[str, typer.Option("--description")]
FlagDescriptionOptional = Annotated[Optional[str], typer.Option("--description")]
FlagUrlOptional = Annotated[Optional[str], typer.Option("--url")]
FlagUrlImageOptional = Annotated[Optional[str], typer.Option("--url-image")]

FlagPublic = Annotated[bool, typer.Option("--public/--private")]
FlagPublicOptional = Annotated[Optional[bool], typer.Option("--public/--private")]
FlagForce = Annotated[bool, typer.Option("--force/--no-force")]
FlagKindRecurse: TypeAlias = Annotated[KindRecurse, typer.Option("--recurse-strategy")]

# FlagContentOptional: TypeAlias = Annotated[Optional[str], typer.Option("--content")]
# FlagMessageOptional: TypeAlias = Annotated[Optional[str], typer.Option("--message")]
# FlagContent: TypeAlias = Annotated[str, typer.Option("--content")]
# FlagMessage: TypeAlias = Annotated[str, typer.Option("--message")]

# --------------------------------------------------------------------------- #
# Search flags


FlagNameLike: TypeAlias = Annotated[Optional[str], typer.Option("--name-like")]
FlagDescriptionLike: TypeAlias = Annotated[
    Optional[str], typer.Option("--description-like")
]
FlagIncludePublic: TypeAlias = Annotated[int, typer.Option("--include-public")]
FlagLimit: TypeAlias = Annotated[int, typer.Option("--limit")]
FlagLimitOptional: TypeAlias = Annotated[Optional[int], typer.Option("--limit")]


# --------------------------------------------------------------------------- #
# Children flags

FlagChildrenUser: TypeAlias = Annotated[
    Optional[ChildrenUser], typer.Option("--child", "--children")
]
FlagChildrenCollection: TypeAlias = Annotated[
    ChildrenCollection, typer.Option("--child", "--children")
]
FlagChildrenDocument: TypeAlias = Annotated[
    ChildrenDocument, typer.Option("--child", "--children")
]


# --------------------------------------------------------------------------- #
# Events flags

ArgUUIDEvent: TypeAlias = Annotated[str, typer.Argument()]
FlagFlatten: TypeAlias = Annotated[
    bool,
    typer.Option(
        "--flatten/--heads",
        help="Return only heads (entries with no parents).",
    ),
]
FlagKind: TypeAlias = Annotated[
    Optional[KindEvent],
    typer.Option("--kind", "-k", help="Matches against the `kind` field."),
]

ArgKindObject: TypeAlias = Annotated[
    KindObject,
    typer.Argument(help="Object kind, matches again the `kind_obj` field."),
]
FlagKindObject: TypeAlias = Annotated[
    KindObject,
    typer.Option(
        "--object",
        "-j",
        help="Object kind, matches again the `kind_obj` field.",
    ),
]
FlagKindObjectOptional: TypeAlias = Annotated[
    Optional[KindObject],
    typer.Option(
        "--object",
        "-j",
        help="Object kind, matches again the `kind_obj` field.",
    ),
]

ArgUUIDEventObject: TypeAlias = Annotated[
    str, typer.Argument(help="Target object uuid.")
]

FlagUUIDEventObject: TypeAlias = Annotated[
    str,
    typer.Option(
        "--uuid-object",
        "--uuid",
        help="Target object UUID.",
    ),
]
FlagUUIDEventObjectOptional: TypeAlias = Annotated[
    Optional[str],
    typer.Option(
        "--uuid-object",
        "--uuid",
        help="Target object UUID.",
    ),
]
FlagBefore = Annotated[Optional[datetime], typer.Option("--before")]
FlagAfter = Annotated[Optional[datetime], typer.Option("--after")]

# --------------------------------------------------------------------------- #
# Configuration Flags.

FlagConfig: TypeAlias = Annotated[
    Optional[str],
    typer.Option("--config", help="Configuration to use everywhere."),
]

FlagHost: TypeAlias = Annotated[
    Optional[str],
    typer.Option(
        "--host",
        help="Host from configuration to use.",
    ),
]
FlagProfile: TypeAlias = Annotated[
    Optional[str],
    typer.Option(
        "--profile",
        "-p",
        help="Profile from configuration to use.",
    ),
]

# --------------------------------------------------------------------------- #
# Loader flags.

ArgFilePath: TypeAlias = Annotated[
    pathlib.Path,
    typer.Argument(help="Path to the apply file."),
]

# --------------------------------------------------------------------------- #
# Token flags

FlagTokenOptional: TypeAlias = Annotated[
    Optional[str], typer.Option(help="Specifies the token to use for any request.")
]

FlagAdmin: TypeAlias = Annotated[
    Optional[bool],
    typer.Option("--admin/--not-an-admin", help="Admin token or not an admin token"),
]

ArgTokenPayload: TypeAlias = Annotated[
    str, typer.Option("--payload", help="Data for token payload.")
]

FlagPending: TypeAlias = Annotated[
    bool,
    typer.Option("--pending", help="Get only pending grants."),
]

FlagOpenApi: TypeAlias = Annotated[
    bool,
    typer.Option(
        "--openapi", help="Get the OpenAPI information for the specified request."
    ),
]

FlagShowRequest: TypeAlias = Annotated[
    bool,
    typer.Option(
        "--request-show",
        help="Show the full request (still executes request).",
    ),
]

FlagPendingFromOptional: TypeAlias = Annotated[
    Optional[PendingFromStr],
    typer.Option("--pending-from", help="Filter results by their pending from status."),
]

FlagNoAuthorization: TypeAlias = Annotated[
    bool,
    typer.Option(
        "--auth-exclude/--auth-include",
        help="Include or exclude the authorization header.",
    ),
]


def check_content(v: str, allow_none: bool = False):
    if allow_none and v is None:
        return v

    decoded = json.loads(v)
    if allow_none and decoded is None:
        return v

    assert isinstance(decoded, dict)
    return decoded


FlagContent: TypeAlias = Annotated[
    str,
    typer.Option(
        "--content",
        help="Content.",
        callback=check_content,
    ),
    Doc("Ignore the string type hint! Should be a `dict`."),
]
FlagContentOptional: TypeAlias = Annotated[
    Optional[str],
    typer.Option(
        "--content",
        help="Content.",
        callback=lambda v: check_content(v, allow_none=True),
    ),
    Doc("Ignore the string type hint! Should be a `dict`."),
]
FlagEmail: TypeAlias = Annotated[
    Optional[str], typer.Option("--email", help="Email address.")
]

FlagDecorate = Annotated[Optional[bool], typer.Option("--decorate/--no-decorate")]
# FlagExclude - Annotated[Optional[List[str]], typer.Option



# =========================================================================== #
from http import HTTPMethod
from random import shuffle
from typing import Any, Dict, List, Type, TypeVar, overload

from fastapi import HTTPException
from sqlalchemy import func
from sqlalchemy.orm import Session

# --------------------------------------------------------------------------- #
# from captura import util
from captura.auth import Token
from captura.controllers.access import Access
from captura.controllers.base import BaseController
from captura.fields import Singular
from captura.models import Collection, Document, Tables, User
from captura.schemas import (
    CollectionSearchSchema,
    DocumentSearchSchema,
    UserSearchSchema,
)

T_ReadParam = TypeVar(
    "T_ReadParam",
)


class Read(BaseController):
    """DO NOT USE THIS FOR GET BY ID, USING :class:`Access` SHOULD BE ENOUGH.

    This is for more complicated methods and will not follow the pattern set
    by :class:`Access`, :class:`Delete`, :class:`Update`, or :class:`Create`.
    """

    access: Access

    def __init__(
        self,
        session: Session,
        token: Token | Dict[str, Any] | None,
        method: HTTPMethod | str,
        *,
        access: Access | None = None,
    ):
        if method != HTTPMethod.GET:
            raise ValueError("`method` must be `GET`.")

        super().__init__(session, token, method)
        if access is None:
            access = self.then(Access)
        self.access = access

    # ======================================================================= #

    @overload
    def search_user(
        self,
        user: User,
        param: UserSearchSchema,
    ) -> List[User]: ...

    @overload
    def search_user(
        self,
        user: User,
        param: DocumentSearchSchema,
    ) -> List[Document]: ...

    @overload
    def search_user(
        self,
        user: User,
        param: CollectionSearchSchema,
    ) -> List[Collection]: ...

    def search_user(
        self,
        user: User,
        param: UserSearchSchema | DocumentSearchSchema | CollectionSearchSchema,
    ) -> List[User] | List[Document] | List[Collection]:
        if param.kind_mapped is None:
            raise HTTPException(500)

        singular = Singular(param.kind_mapped.name)

        T_kind: Type[User] | Type[Document] | Type[Collection]
        T_kind = Tables[singular.name].value  # type: ignore[reportGeneralTypeErrors]

        q = T_kind.q_search(
            user.uuid,
            param.uuids,
            all_=True,
            name_like=param.name_like,
            description_like=param.description_like,
            limit=param.limit,
        )

        if param.randomize and param.uuids is None:
            # if param.limit is None or param.limit > 25:
            #     msg = "Limit must be less than `25` to randomize."
            #     raise HTTPException(400, detail=msg)

            q = q.order_by(func.random())

        res: List[User] | List[Document] | List[Collection]
        res = list(self.session.scalars(q))

        if param.randomize and param.uuids is not None:
            shuffle(res)

        return res
